from unittest.mock import Mock, patch

import pytest
from napari_plugin_engine import napari_hook_implementation
from qtpy.QtWidgets import QWidget

import napari
from napari import Viewer
from napari._qt.menus import PluginsMenu
from napari._qt.qt_main_window import _instantiate_dock_widget
from napari.utils._proxies import PublicOnlyProxy


class Widg1(QWidget):
    pass


class Widg2(QWidget):
    def __init__(self, napari_viewer):
        self.viewer = napari_viewer
        super().__init__()


class Widg3(QWidget):
    def __init__(self, v: Viewer):
        self.viewer = v
        super().__init__()

    def fail(self):
        """private attr not allowed"""
        self.viewer.window._qt_window


def magicfunc(viewer: 'napari.Viewer'):
    return viewer


dwidget_args = {
    'single_class': Widg1,
    'class_tuple': (Widg1, {'area': 'right'}),
    'tuple_list': [(Widg1, {'area': 'right'}), (Widg2, {})],
    'tuple_list2': [(Widg1, {'area': 'right'}), Widg2],
    'bad_class': 1,
    'bad_tuple1': (Widg1, 1),
    'bad_double_tuple': ((Widg1, {}), (Widg2, {})),
}


# napari_plugin_manager from _testsupport.py
# monkeypatch, request, recwarn fixtures are from pytest
@pytest.mark.parametrize('arg', dwidget_args.values(), ids=dwidget_args.keys())
def test_dock_widget_registration(
    arg, napari_plugin_manager, request, recwarn
):
    """Test that dock widgets get validated and registerd correctly."""

    class Plugin:
        @napari_hook_implementation
        def napari_experimental_provide_dock_widget():
            return arg

    napari_plugin_manager.register(Plugin, name='Plugin')
    napari_plugin_manager.discover_widgets()
    widgets = napari_plugin_manager._dock_widgets

    if '[bad_' in request.node.name:
        assert len(recwarn) == 1
        assert not widgets
    else:
        assert len(recwarn) == 0
        assert widgets['Plugin']['Widg1'][0] == Widg1
        if 'tuple_list' in request.node.name:
            assert widgets['Plugin']['Widg2'][0] == Widg2


@pytest.fixture
def test_plugin_widgets(monkeypatch, napari_plugin_manager):
    """A smattering of example registered dock widgets and function widgets."""
    tnpm = napari_plugin_manager
    dock_widgets = {
        "TestP1": {"Widg1": (Widg1, {}), "Widg2": (Widg2, {})},
        "TestP2": {"Widg3": (Widg3, {})},
    }
    monkeypatch.setattr(tnpm, "_dock_widgets", dock_widgets)

    function_widgets = {'TestP3': {'magic': magicfunc}}
    monkeypatch.setattr(tnpm, "_function_widgets", function_widgets)
    yield


def test_plugin_widgets_menus(test_plugin_widgets, qtbot):
    """Test the plugin widgets get added to the window menu correctly."""
    # only take the plugin actions
    window = Mock()
    qtwin = QWidget()
    qtbot.addWidget(qtwin)
    with patch.object(window, '_qt_window', qtwin):
        actions = PluginsMenu(window=window).actions()
    for cnt, action in enumerate(actions):
        if action.text() == "":
            break
    actions = actions[cnt + 1 :]
    texts = [a.text() for a in actions]
    for t in ['TestP1', 'Widg3 (TestP2)', 'magic (TestP3)']:
        assert t in texts

    # Expect a submenu ("Test plugin1") with particular entries.
    tp1 = next(m for m in actions if m.text() == 'TestP1')
    assert tp1.parent()
    assert [a.text() for a in tp1.parent().actions()] == ['Widg1', 'Widg2']


def test_making_plugin_dock_widgets(test_plugin_widgets, make_napari_viewer):
    """Test that we can create dock widgets, and they get the viewer."""
    viewer = make_napari_viewer()
    # only take the plugin actions
    actions = viewer.window.plugins_menu.actions()
    for cnt, action in enumerate(actions):
        if action.text() == "":
            # this is the separator
            break
    actions = actions[cnt + 1 :]

    # trigger the 'TestP2: Widg3' action
    tp2 = next(m for m in actions if m.text().endswith('(TestP2)'))
    tp2.trigger()
    # make sure that a dock widget was created
    assert 'Widg3 (TestP2)' in viewer.window._dock_widgets
    dw = viewer.window._dock_widgets['Widg3 (TestP2)']
    assert isinstance(dw.widget(), Widg3)
    # This widget uses the parameter annotation method to receive a viewer
    assert isinstance(dw.widget().viewer, napari.Viewer)
    # Add twice is ok, only does a show
    tp2.trigger()

    # trigger the 'TestP1 > Widg2' action (it's in a submenu)
    tp2 = next(m for m in actions if m.text().endswith('TestP1'))
    action = tp2.parent().actions()[1]
    assert action.text() == 'Widg2'
    action.trigger()
    # make sure that a dock widget was created
    assert 'Widg2 (TestP1)' in viewer.window._dock_widgets
    dw = viewer.window._dock_widgets['Widg2 (TestP1)']
    assert isinstance(dw.widget(), Widg2)
    # This widget uses parameter *name* "napari_viewer" to get a viewer
    assert isinstance(dw.widget().viewer, napari.Viewer)
    # Add twice is ok, only does a show
    action.trigger()
    # Check that widget is still there when closed.
    widg = dw.widget()
    dw.title.hide_button.click()
    assert widg
    # Check that widget is destroyed when closed.
    dw.destroyOnClose()
    assert action not in viewer.window.plugins_menu.actions()
    assert not widg.parent()


def test_making_function_dock_widgets(test_plugin_widgets, make_napari_viewer):
    """Test that we can create magicgui widgets, and they get the viewer."""
    import magicgui

    viewer = make_napari_viewer()
    # only take the plugin actions
    actions = viewer.window.plugins_menu.actions()
    for cnt, action in enumerate(actions):
        if action.text() == "":
            # this is the separator
            break
    actions = actions[cnt + 1 :]

    # trigger the 'TestP3: magic' action
    tp3 = next(m for m in actions if m.text().endswith('(TestP3)'))
    tp3.trigger()
    # make sure that a dock widget was created
    assert 'magic (TestP3)' in viewer.window._dock_widgets
    dw = viewer.window._dock_widgets['magic (TestP3)']
    # make sure that it contains a magicgui widget
    magic_widget = dw.widget()._magic_widget
    FGui = getattr(magicgui.widgets, 'FunctionGui', None)
    if FGui is None:
        # pre magicgui 0.2.6
        FGui = magicgui.FunctionGui
    assert isinstance(magic_widget, FGui)
    # This magicgui widget uses the parameter annotation to receive a viewer
    assert isinstance(magic_widget.viewer.value, napari.Viewer)
    # The function just returns the viewer... make sure we can call it
    assert isinstance(magic_widget(), napari.Viewer)
    # Add twice is ok, only does a show
    tp3.trigger()


def test_inject_viewer_proxy(make_napari_viewer):
    """Test that the injected viewer is a public-only proxy"""
    viewer = make_napari_viewer()
    wdg = _instantiate_dock_widget(Widg3, viewer)
    assert isinstance(wdg.viewer, PublicOnlyProxy)

    # simulate access from outside napari
    with patch('napari.utils.misc.ROOT_DIR', new='/some/other/package'):
        with pytest.warns(FutureWarning):
            wdg.fail()



import dask.array as da
import numpy as np
import pytest

from napari.components import ViewerModel
from napari.utils.colormaps import (
    AVAILABLE_COLORMAPS,
    CYMRGB,
    MAGENTA_GREEN,
    SIMPLE_COLORMAPS,
    Colormap,
    ensure_colormap,
)
from napari.utils.misc import ensure_iterable, ensure_sequence_of_iterables

base_colormaps = CYMRGB
two_colormaps = MAGENTA_GREEN
green_cmap = SIMPLE_COLORMAPS['green']
red_cmap = SIMPLE_COLORMAPS['red']
blue_cmap = AVAILABLE_COLORMAPS['blue']
cmap_tuple = ("my_colormap", Colormap(['g', 'm', 'y']))
cmap_dict = {"your_colormap": Colormap(['g', 'r', 'y'])}

MULTI_TUPLES = [[0.3, 0.7], [0.1, 0.9], [0.3, 0.9], [0.4, 0.9], [0.2, 0.9]]

# data shape is (15, 10, 5) unless otherwise set
# channel_axis = -1 is implied unless otherwise set
multi_channel_test_data = [
    # basic multichannel image
    ((), {}),
    # single channel
    ((15, 10, 1), {}),
    # two channels
    ((15, 10, 2), {}),
    # Test adding multichannel image with color channel set.
    ((5, 10, 15), {'channel_axis': 0}),
    # split single RGB image
    ((15, 10, 3), {'colormap': ['red', 'green', 'blue']}),
    # multiple RGB images
    ((15, 10, 5, 3), {'channel_axis': 2, 'rgb': True}),
    # Test adding multichannel image with custom names.
    ((), {'name': ['multi ' + str(i + 3) for i in range(5)]}),
    # Test adding multichannel image with custom contrast limits.
    ((), {'contrast_limits': [0.3, 0.7]}),
    ((), {'contrast_limits': MULTI_TUPLES}),
    ((), {'gamma': 0.5}),
    ((), {'gamma': [0.3, 0.4, 0.5, 0.6, 0.7]}),
    ((), {'visible': [True, False, False, True, True]}),
    # Test adding multichannel image with custom colormaps.
    ((), {'colormap': 'gray'}),
    ((), {'colormap': green_cmap}),
    ((), {'colormap': cmap_tuple}),
    ((), {'colormap': cmap_dict}),
    ((), {'colormap': ['gray', 'blue', 'red', 'green', 'yellow']}),
    (
        (),
        {'colormap': [green_cmap, red_cmap, blue_cmap, blue_cmap, green_cmap]},
    ),
    ((), {'colormap': [green_cmap, 'gray', cmap_tuple, blue_cmap, cmap_dict]}),
    ((), {'scale': MULTI_TUPLES}),
    ((), {'translate': MULTI_TUPLES}),
    ((), {'blending': 'translucent'}),
    ((), {'metadata': {'hi': 'there'}}),
    ((), {'metadata': {k: v for k, v in MULTI_TUPLES}}),
    ((), {'experimental_clipping_planes': []}),
]

ids = [
    'basic_multichannel',
    'one_channel',
    'two_channel',
    'specified_multichannel',
    'split_RGB',
    'list_RGB',
    'names',
    'contrast_limits_broadcast',
    'contrast_limits_list',
    'gamma_broadcast',
    'gamma_list',
    'visibility',
    'colormap_string_broadcast',
    'colormap_cmap_broadcast',
    'colormap_tuple_broadcast',
    'colormap_dict_broadcast',
    'colormap_string_list',
    'colormap_cmap_list',
    'colormap_variable_list',
    'scale',
    'translate',
    'blending',
    'metadata_broadcast',
    'metadata_multi',
    'empty_clipping_planes',
]


@pytest.mark.parametrize('shape, kwargs', multi_channel_test_data, ids=ids)
def test_multichannel(shape, kwargs):
    """Test adding multichannel image."""
    viewer = ViewerModel()
    np.random.seed(0)
    data = np.random.random(shape or (15, 10, 5))
    channel_axis = kwargs.pop('channel_axis', -1)
    viewer.add_image(data, channel_axis=channel_axis, **kwargs)

    # make sure the right number of layers got added
    n_channels = data.shape[channel_axis]
    assert len(viewer.layers) == n_channels

    for i in range(n_channels):
        # make sure that the data has been divided into layers
        assert np.all(viewer.layers[i].data == data.take(i, axis=channel_axis))
        # make sure colors have been assigned properly
        if 'colormap' not in kwargs:
            if n_channels == 1:
                assert viewer.layers[i].colormap.name == 'gray'
            elif n_channels == 2:
                assert viewer.layers[i].colormap.name == two_colormaps[i]
            else:
                assert viewer.layers[i].colormap.name == base_colormaps[i]
        if 'blending' not in kwargs:
            assert (
                viewer.layers[i].blending == 'translucent_no_depth'
                if i == 0
                else 'additive'
            )
        for key, expectation in kwargs.items():
            # broadcast exceptions
            if key in {
                'scale',
                'translate',
                'rotate',
                'shear',
                'contrast_limits',
                'metadata',
                'experimental_clipping_planes',
            }:
                expectation = ensure_sequence_of_iterables(
                    expectation, repeat_empty=True
                )
            elif key == 'colormap' and expectation is not None:
                if isinstance(expectation, list):
                    exp = [ensure_colormap(c).name for c in expectation]
                else:
                    exp = ensure_colormap(expectation).name
                expectation = ensure_iterable(exp)
            else:
                expectation = ensure_iterable(expectation)
            expectation = [v for i, v in zip(range(i + 1), expectation)]

            result = getattr(viewer.layers[i], key)
            if key == 'colormap':  # colormaps are tuples of (name, cmap)
                result = result.name
            if isinstance(result, np.ndarray):
                np.testing.assert_almost_equal(result, expectation[i])
            else:
                assert result == expectation[i]


def test_multichannel_multiscale():
    """Test adding multichannel multiscale."""
    viewer = ViewerModel()
    np.random.seed(0)
    shapes = [(40, 20, 4), (20, 10, 4), (10, 5, 4)]
    np.random.seed(0)
    data = [np.random.random(s) for s in shapes]
    viewer.add_image(data, channel_axis=-1, multiscale=True)
    assert len(viewer.layers) == data[0].shape[-1]
    for i in range(data[0].shape[-1]):
        assert np.all(
            [
                np.all(l_d == d)
                for l_d, d in zip(
                    viewer.layers[i].data,
                    [data[j].take(i, axis=-1) for j in range(len(data))],
                )
            ]
        )
        assert viewer.layers[i].colormap.name == base_colormaps[i]


def test_multichannel_implicit_multiscale():
    """Test adding multichannel implicit multiscale."""
    viewer = ViewerModel()
    np.random.seed(0)
    shapes = [(40, 20, 4), (20, 10, 4), (10, 5, 4)]
    np.random.seed(0)
    data = [np.random.random(s) for s in shapes]
    viewer.add_image(data, channel_axis=-1)
    assert len(viewer.layers) == data[0].shape[-1]
    for i in range(data[0].shape[-1]):
        assert np.all(
            [
                np.all(l_d == d)
                for l_d, d in zip(
                    viewer.layers[i].data,
                    [data[j].take(i, axis=-1) for j in range(len(data))],
                )
            ]
        )
        assert viewer.layers[i].colormap.name == base_colormaps[i]


def test_multichannel_dask_array():
    """Test adding multichannel dask array."""
    viewer = ViewerModel()
    np.random.seed(0)
    data = da.random.random((2, 10, 10, 5))
    viewer.add_image(data, channel_axis=0)
    assert len(viewer.layers) == data.shape[0]
    for i in range(data.shape[0]):
        assert viewer.layers[i].data.shape == data.shape[1:]
        assert isinstance(viewer.layers[i].data, type(data))


def test_forgot_multichannel_error_hint():
    """Test that a helpful error is raised when channel_axis is not used."""
    viewer = ViewerModel()
    np.random.seed(0)
    data = da.random.random((15, 10, 5))
    with pytest.raises(TypeError) as e:
        viewer.add_image(data, name=['a', 'b', 'c'])
    assert "did you mean to specify a 'channel_axis'" in str(e)


def test_multichannel_index_error_hint():
    """Test multichannel error when arg length != n_channels."""
    viewer = ViewerModel()
    np.random.seed(0)
    data = da.random.random((5, 10, 5))
    with pytest.raises(IndexError) as e:
        viewer.add_image(data, channel_axis=0, name=['a', 'b'])
    assert (
        "Requested channel_axis (0) had length 5, but the "
        "'name' argument only provided 2 values." in str(e)
    )



from superqt.utils._qthreading import (
    GeneratorWorkerSignals,
    WorkerBase,
    WorkerBaseSignals,
)

from .._qt.qthreading import (
    FunctionWorker,
    GeneratorWorker,
    create_worker,
    thread_worker,
)

# all of these might be used by an end-user when subclassing
__all__ = (
    'create_worker',
    'FunctionWorker',
    'GeneratorWorker',
    'GeneratorWorkerSignals',
    'thread_worker',
    'WorkerBase',
    'WorkerBaseSignals',
)



from unittest.mock import MagicMock

from qtpy.QtWidgets import QMenu

from napari._qt.menus._util import populate_menu


def test_populate_menu_create(qtbot):
    """Test the populate_menu function."""

    mock = MagicMock()
    menu = QMenu()
    populate_menu(menu, [{"text": "test", "slot": mock}])
    assert len(menu.actions()) == 1
    assert menu.actions()[0].text() == "test"
    assert menu.actions()[0].isCheckable() is False
    with qtbot.waitSignal(menu.actions()[0].triggered):
        menu.actions()[0].trigger()
    mock.assert_called_once()


def test_populate_menu_create_checkable(qtbot):
    """Test the populate_menu function with checkable actions."""

    mock = MagicMock()
    menu = QMenu()
    populate_menu(menu, [{"text": "test", "slot": mock, "checkable": True}])
    assert len(menu.actions()) == 1
    assert menu.actions()[0].text() == "test"
    assert menu.actions()[0].isCheckable() is True
    with qtbot.waitSignal(menu.actions()[0].triggered):
        menu.actions()[0].trigger()
    mock.assert_called_once_with(True)
    mock.reset_mock()
    with qtbot.waitSignal(menu.actions()[0].triggered):
        menu.actions()[0].trigger()
    mock.assert_called_once_with(False)



"""Scale bar model."""
from typing import Optional

from napari.utils.color import ColorValue

from ..utils.events import EventedModel
from ._viewer_constants import Position


class ScaleBar(EventedModel):
    """Scale bar indicating size in world coordinates.

    Attributes
    ----------
    visible : bool
        If scale bar is visible or not.
    colored : bool
        If scale bar are colored or not. If colored then
        default color is magenta. If not colored than
        scale bar color is the opposite of the canvas
        background or the background box.
    color : ColorValue
        Scalebar and text color.
        See ``ColorValue.validate`` for supported values.
    ticks : bool
        If scale bar has ticks at ends or not.
    position : str
        Position of the scale bar in the canvas. Must be one of
        'top left', 'top right', 'bottom right', 'bottom left'.
        Default value is 'bottom right'.
    background_color : np.ndarray
        Background color of canvas. If scale bar is not colored
        then it has the color opposite of this color.
    font_size : float
        The font size (in points) of the text.
    box : bool
        If background box is visible or not.
    box_color : Optional[str | array-like]
        Background box color.
        See ``ColorValue.validate`` for supported values.
    unit : Optional[str]
        Unit to be used by the scale bar. The value can be set
        to `None` to display no units.
    """

    visible: bool = False
    colored: bool = False
    color: ColorValue = [1, 0, 1, 1]
    ticks: bool = True
    position: Position = Position.BOTTOM_RIGHT
    font_size: float = 10
    box: bool = False
    box_color: ColorValue = [0, 0, 0, 0.6]
    unit: Optional[str] = None



"""This module contains functions that 'standardize' the color handling
of napari layers by supplying functions that are able to convert most
color representation the user had in mind into a single representation -
a numpy Nx4 array of float32 values between 0 and 1 - that is used across
the codebase. The color is always in an RGBA format. To handle colors in
HSV, for example, we should point users to skimage, matplotlib and others.

The main function of the module is "transform_color", which might call
a cascade of other, private, function in the module to do the hard work
of converting the input. This function will either be called directly, or
used by the function "transform_color_with_defaults", which is a helper
function for the layer objects located in
``layers.utils.color_transformations.py``.

In general, when handling colors we try to catch invalid color
representations, warn the users of their misbehaving and return a default
white color array, since it seems unreasonable to crash the entire napari
session due to mis-represented colors.
"""

import functools
import types
import warnings
from typing import Any, Callable, Dict, Sequence

import numpy as np
from vispy.color import ColorArray, get_color_dict, get_color_names
from vispy.color.color_array import _string_to_rgb

from ..translations import trans


def transform_color(colors: Any) -> np.ndarray:
    """Transforms provided color(s) to an Nx4 array of RGBA np.float32
    values.

    N is the number of given colors. The function is designed to parse all
    valid color representations a user might have and convert them properly.
    That being said, combinations of different color representation in the
    same list of colors is prohibited, and will error. This means that a list
    of ['red', np.array([1, 0, 0])] cannot be parsed and has to be manually
    pre-processed by the user before sent to this function. In addition, the
    provided colors - if numeric - should already be in an RGB(A) format. To
    convert an existing numeric color array to RGBA format use skimage before
    calling this function.

    Parameters
    ----------
    colors : string and array-like.
        The color(s) to interpret and convert

    Returns
    -------
    colors : np.ndarray
        An instance of np.ndarray with a data type of float32, 4 columns in
        RGBA order and N rows, with N being the number of colors. The array
        will always be 2D even if a single color is passed.

    Raises
    ------
    ValueError, AttributeError, KeyError
        invalid inputs
    """
    colortype = type(colors)
    return _color_switch[colortype](colors)


@functools.lru_cache(maxsize=1024)
def _handle_str(color: str) -> np.ndarray:
    """Creates an array from a color of type string.

    The function uses an LRU cache to enhance performance.

    Parameters
    ----------
    color : str
        A single string as an input color. Can be a color name or a
        hex representation of a color, with either 6 or 8 hex digits.

    Returns
    -------
    colorarray : np.ndarray
        1x4 array

    """
    if len(color) == 0:
        warnings.warn(
            trans._(
                "Empty string detected. Returning black instead.",
                deferred=True,
            )
        )
        return np.zeros((1, 4), dtype=np.float32)

    colorarray = np.atleast_2d(_string_to_rgb(color)).astype(np.float32)
    if colorarray.shape[1] == 3:
        colorarray = np.column_stack([colorarray, np.float32(1.0)])
    return colorarray


def _handle_list_like(colors: Sequence) -> np.ndarray:
    """Parse a list-like container of colors into a numpy Nx4 array.

    Handles all list-like containers of colors using recursion (if necessary).
    The colors inside the container should all be represented in the same
    manner. This means that a list containing ['r', (1., 1., 1.)] will raise
    an error. Note that numpy arrays are handled in _handle_array. Lists which
    are known to contain strings will be parsed with _handle_str_list_like.
    Generators should first visit _handle_generator before arriving as input.

    Parameters
    ----------
    colors : Sequence
        A list-like container of colors. The colors inside should be homogeneous
        in their representation.

    Returns
    -------
    color_array : np.ndarray
        Nx4 numpy array, with N being the length of ``colors``.
    """

    try:
        # The following conversion works for most cases, and so it's expected
        # that most valid inputs will pass this .asarray() call
        # with ease. Those who don't are usually too cryptic to decipher.
        # If only some of the colors are strings, explicitly provide an object
        # dtype to avoid the deprecated behavior described in:
        # https://github.com/napari/napari/issues/2791
        num_str = len([c for c in colors if isinstance(c, str)])
        dtype = 'O' if 0 < num_str < len(colors) else None
        color_array = np.atleast_2d(np.asarray(colors, dtype=dtype))
    except ValueError:
        warnings.warn(
            trans._(
                "Couldn't convert input color array to a proper numpy array. Please make sure that your input data is in a parsable format. Converting input to a white color array.",
                deferred=True,
            )
        )
        return np.ones((max(len(colors), 1), 4), dtype=np.float32)

    # Happy path - converted to a float\integer array
    if color_array.dtype.kind in ['f', 'i']:
        return _handle_array(color_array)

    # User input was an iterable with strings
    if color_array.dtype.kind in ['U', 'O']:
        return _handle_str_list_like(color_array.ravel())


def _handle_generator(colors) -> np.ndarray:
    """Generators are converted to lists since we need to know their
    length to instantiate a proper array.
    """
    return _handle_list_like(list(colors))


def handle_nested_colors(colors) -> ColorArray:
    """In case of an array-like container holding colors, unpack it."""
    colors_as_rbga = np.ones((len(colors), 4), dtype=np.float32)
    for idx, color in enumerate(colors):
        colors_as_rbga[idx] = _color_switch[type(color)](color)
    return ColorArray(colors_as_rbga)


def _handle_array(colors: np.ndarray) -> np.ndarray:
    """Converts the given array into an array in the right format."""
    kind = colors.dtype.kind

    # Object arrays aren't handled by napari
    if kind == 'O':
        warnings.warn(
            trans._(
                "An object array was passed as the color input. Please convert its datatype before sending it to napari. Converting input to a white color array.",
                deferred=True,
            )
        )
        return np.ones((max(len(colors), 1), 4), dtype=np.float32)

    # An array of strings will be treated as a list if compatible
    elif kind == 'U':
        if colors.ndim == 1:
            return _handle_str_list_like(colors)
        else:
            warnings.warn(
                trans._(
                    "String color arrays should be one-dimensional. Converting input to a white color array.",
                    deferred=True,
                )
            )
            return np.ones((len(colors), 4), dtype=np.float32)

    # Test the dimensionality of the input array

    # Empty color array can be a way for the user to signal
    # that it wants the "default" colors of napari. We return
    # a single white color.
    if colors.shape[-1] == 0:
        warnings.warn(
            trans._(
                "Given color input is empty. Converting input to a white color array.",
                deferred=True,
            )
        )
        return np.ones((1, 4), dtype=np.float32)

    colors = np.atleast_2d(colors)

    # Arrays with more than two dimensions don't have a clear
    # conversion method to a color array and thus raise an error.
    if colors.ndim > 2:
        raise ValueError(
            trans._(
                "Given colors input should contain one or two dimensions. Received array with {ndim} dimensions.",
                deferred=True,
                ndim=colors.ndim,
            )
        )

    # User provided a list of numbers as color input. This input
    # cannot be coerced into something understandable and thus
    # will return an error.
    if colors.shape[0] == 1 and colors.shape[1] not in {3, 4}:
        raise ValueError(
            trans._(
                "Given color array has an unsupported format. Received the following array:\n{colors}\nA proper color array should have 3-4 columns with a row per data entry.",
                deferred=True,
                colors=colors,
            )
        )

    # The user gave a list of colors, but it contains a wrong number
    # of columns. This check will also drop Nx1 (2D) arrays, since
    # numpy has vectors, and representing colors in this way
    # (column vector-like) is redundant. However, this results in a
    # warning and not a ValueError since we know the number of colors
    # in this dataset, meaning we can save the napari session by
    # rendering the data in white, which better than crashing.
    if not 3 <= colors.shape[1] <= 4:
        warnings.warn(
            trans._(
                "Given colors input should contain three or four columns. Received array with {shape} columns. Converting input to a white color array.",
                deferred=True,
                shape=colors.shape[1],
            )
        )
        return np.ones((len(colors), 4), dtype=np.float32)

    # Arrays with floats and ints can be safely converted to the proper format
    if kind in ['f', 'i', 'u']:
        return _convert_array_to_correct_format(colors)

    else:
        raise ValueError(
            trans._(
                "Data type of array ({color_dtype}) not supported.",
                deferred=True,
                color_dtype=colors.dtype,
            )
        )


def _convert_array_to_correct_format(colors: np.ndarray) -> np.ndarray:
    """Asserts shape, dtype and normalization of given color array.

    This function deals with arrays which are already 'well-behaved',
    i.e. have (almost) the correct number of columns and are able to represent
    colors correctly. It then it makes sure that the array indeed has exactly
    four columns and that its values are normalized between 0 and 1, with a
    data type of float32.

    Parameters
    ----------
    colors : np.ndarray
        Input color array, perhaps un-normalized and without the alpha channel.

    Returns
    -------
    colors : np.ndarray
        Nx4, float32 color array with values in the range [0, 1]
    """
    if colors.shape[1] == 3:
        colors = np.column_stack(
            [colors, np.ones(len(colors), dtype=np.float32)]
        )

    if colors.min() < 0:
        raise ValueError(
            trans._(
                "Colors input had negative values.",
                deferred=True,
            )
        )

    if colors.max() > 1:
        warnings.warn(
            trans._(
                "Colors with values larger than one detected. napari will normalize these colors for you. If you'd like to convert these yourself, please use the proper method from skimage.color.",
                deferred=True,
            )
        )
        colors = _normalize_color_array(colors)
    return np.atleast_2d(np.asarray(colors, dtype=np.float32))


def _handle_str_list_like(colors: Sequence) -> np.ndarray:
    """Converts lists or arrays filled with strings to the proper color array
    format.

    Parameters
    ----------
    colors : list-like
        A sequence of string colors

    Returns
    -------
    color_array : np.ndarray
        Nx4, float32 color array
    """
    color_array = np.empty((len(colors), 4), dtype=np.float32)
    for idx, c in enumerate(colors):
        try:
            color_array[idx, :] = _color_switch[type(c)](c)
        except (ValueError, TypeError, KeyError):
            raise ValueError(
                trans._(
                    "Invalid color found: {color} at index {idx}.",
                    deferred=True,
                    color=c,
                    idx=idx,
                )
            )
    return color_array


def _handle_none(color) -> np.ndarray:
    """Converts color given as None to black.

    Parameters
    ----------
    color : NoneType
        None value given as a color

    Returns
    -------
    arr : np.ndarray
        1x4 numpy array of float32 zeros

    """
    return np.zeros((1, 4), dtype=np.float32)


def _normalize_color_array(colors: np.ndarray) -> np.ndarray:
    """Normalize all array values to the range [0, 1].

    The added complexity here stems from the fact that if a row in the given
    array contains four identical value a simple normalization might raise a
    division by zero exception.

    Parameters
    ----------
    colors : np.ndarray
        A numpy array with values possibly outside the range of [0, 1]

    Returns
    -------
    colors : np.ndarray
        Copy of input array with normalized values
    """
    colors = colors.astype(np.float32, copy=True)
    out_of_bounds_idx = np.unique(np.where((colors > 1) | (colors < 0))[0])
    out_of_bounds = colors[out_of_bounds_idx]
    norm = np.linalg.norm(out_of_bounds, np.inf, axis=1)
    out_of_bounds = out_of_bounds / norm[:, np.newaxis]
    colors[out_of_bounds_idx] = out_of_bounds
    return colors.astype(np.float32)


_color_switch: Dict[Any, Callable] = {
    str: _handle_str,
    np.str_: _handle_str,
    list: _handle_list_like,
    tuple: _handle_list_like,
    types.GeneratorType: _handle_generator,
    np.ndarray: _handle_array,
    type(None): _handle_none,
}


def _create_hex_to_name_dict():
    """Create a dictionary mapping hexadecimal RGB colors into their
    'official' name.

    Returns
    -------
    hex_to_rgb : dict
        Mapping from hexadecimal RGB ('#ff0000') to name ('red').
    """
    colordict = get_color_dict()
    hex_to_name = {f"{v.lower()}ff": k for k, v in colordict.items()}
    return hex_to_name


def get_color_namelist():
    """Gets all the color names supported by napari.

    Returns
    -------
    list[str]
        All the color names supported by napari.
    """
    return get_color_names()


hex_to_name = _create_hex_to_name_dict()


def _check_color_dim(val):
    """Ensures input is Nx4.

    Parameters
    ----------
    val : np.ndarray
        A color array of possibly less than 4 columns

    Returns
    -------
    val : np.ndarray
        A four columns version of the input array. If the original array
        was a missing the fourth channel, it's added as 1.0 values.
    """
    val = np.atleast_2d(val)
    if val.shape[1] not in (3, 4):
        strval = str(val)
        if len(strval) > 100:
            strval = strval[:97] + '...'
        raise RuntimeError(
            trans._(
                'Value must have second dimension of size 3 or 4. Got `{val}`, shape={shape}',
                deferred=True,
                shape=val.shape,
                val=strval,
            )
        )

    if val.shape[1] == 3:
        val = np.column_stack([val, np.float32(1.0)])
    return val


def rgb_to_hex(rgbs: Sequence) -> np.ndarray:
    """Convert RGB to hex quadruplet.

    Taken from vispy with slight modifications.

    Parameters
    ----------
    rgbs : Sequence
        A list-like container of colors in RGBA format with values
        between [0, 1]

    Returns
    -------
    arr : np.ndarray
        An array of the hex representation of the input colors

    """
    rgbs = _check_color_dim(rgbs)
    return np.array(
        [
            f'#{"%02x" * 4}' % tuple((255 * rgb).astype(np.uint8))
            for rgb in rgbs
        ],
        '|U9',
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy.QtCore import QSortFilterProxyModel, Qt

from ...layers import Layer
from ...utils.translations import trans
from ._base_item_model import SortRole, _BaseEventedItemModel
from ._layer_delegate import LayerDelegate
from .qt_list_view import QtListView

if TYPE_CHECKING:
    from qtpy.QtGui import QKeyEvent
    from qtpy.QtWidgets import QWidget

    from ...components.layerlist import LayerList


class ReverseProxyModel(QSortFilterProxyModel):
    """Proxy Model that reverses the view order of a _BaseEventedItemModel."""

    def __init__(self, model: _BaseEventedItemModel) -> None:
        super().__init__()
        self.setSourceModel(model)
        self.setSortRole(SortRole)
        self.sort(0, Qt.SortOrder.DescendingOrder)

    def dropMimeData(self, data, action, destRow, col, parent):
        """Handle destination row for dropping with reversed indices."""
        row = 0 if destRow == -1 else self.sourceModel().rowCount() - destRow
        return self.sourceModel().dropMimeData(data, action, row, col, parent)


class QtLayerList(QtListView[Layer]):
    """QItemView subclass specialized for the LayerList.

    This is as mostly for targetting with QSS, applying the delegate and
    reversing the view with ReverseProxyModel.
    """

    def __init__(self, root: LayerList, parent: QWidget = None):
        super().__init__(root, parent)
        self.setItemDelegate(LayerDelegate())
        self.setToolTip(trans._('Layer list'))
        font = self.font()
        font.setPointSize(12)
        self.setFont(font)

        # This reverses the order of the items in the view,
        # so items at the end of the list are at the top.
        self.setModel(ReverseProxyModel(self.model()))

    def keyPressEvent(self, e: QKeyEvent) -> None:
        """Override Qt event to pass events to the viewer."""
        if e.key() != Qt.Key.Key_Space:
            super().keyPressEvent(e)
        if e.key() not in (Qt.Key.Key_Backspace, Qt.Key.Key_Delete):
            e.ignore()  # pass key events up to viewer



import warnings
from functools import wraps

from napari.utils.translations import trans


def rename_argument(from_name: str, to_name: str, version: str):
    """
    This is decorator for simple rename function argument
    without break backward compatibility.

    Parameters
    ----------
    from_name : str
        old name of argument
    to_name : str
        new name of argument
    """

    def _wrapper(func):
        @wraps(func)
        def _update_from_dict(*args, **kwargs):
            if from_name in kwargs:
                if to_name in kwargs:
                    raise ValueError(
                        trans._(
                            "Argument {to_name} already defined, please do not mix {from_name} and {to_name} in one call.",
                            from_name=from_name,
                            to_name=to_name,
                        )
                    )
                warnings.warn(
                    trans._(
                        "Argument {from_name} is deprecated, please use {to_name} instead. It will be removed in {version}.",
                        from_name=from_name,
                        to_name=to_name,
                        version=version,
                    ),
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                kwargs = kwargs.copy()
                kwargs[to_name] = kwargs.pop(from_name)
            return func(*args, **kwargs)

        return _update_from_dict

    return _wrapper



from __future__ import annotations

import itertools
import os.path
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict, namedtuple
from contextlib import contextmanager
from functools import cached_property
from typing import List, Optional, Tuple, Union

import magicgui as mgui
import numpy as np
from npe2 import plugin_manager as pm

from ...utils._dask_utils import configure_dask
from ...utils._magicgui import (
    add_layer_to_viewer,
    add_layers_to_viewer,
    get_layers,
)
from ...utils.events import EmitterGroup, Event
from ...utils.events.event import WarningEmitter
from ...utils.geometry import (
    find_front_back_face,
    intersect_line_with_axis_aligned_bounding_box_3d,
)
from ...utils.key_bindings import KeymapProvider
from ...utils.mouse_bindings import MousemapProvider
from ...utils.naming import magic_name
from ...utils.status_messages import generate_layer_coords_status
from ...utils.transforms import Affine, CompositeAffine, TransformChain
from ...utils.translations import trans
from ..utils.interactivity_utils import drag_data_to_projected_distance
from ..utils.layer_utils import (
    coerce_affine,
    compute_multiscale_level_and_corners,
    convert_to_uint8,
    dims_displayed_world_to_layer,
    get_extent_world,
)
from ..utils.plane import ClippingPlane, ClippingPlaneList
from ._base_constants import Blending

Extent = namedtuple('Extent', 'data world step')


def no_op(layer: Layer, event: Event) -> None:
    """
    A convenient no-op event for the layer mouse binding.

    This makes it easier to handle many cases by inserting this as
    as place holder

    Parameters
    ----------
    layer : Layer
        Current layer on which this will be bound as a callback
    event : Event
        event that triggered this mouse callback.

    Returns
    -------
    None

    """
    return None


@mgui.register_type(choices=get_layers, return_callback=add_layer_to_viewer)
class Layer(KeymapProvider, MousemapProvider, ABC):
    """Base layer class.

    Parameters
    ----------
    name : str
        Name of the layer.
    metadata : dict
        Layer metadata.
    scale : tuple of float
        Scale factors for the layer.
    translate : tuple of float
        Translation values for the layer.
    rotate : float, 3-tuple of float, or n-D array.
        If a float convert into a 2D rotation matrix using that value as an
        angle. If 3-tuple convert into a 3D rotation matrix, using a yaw,
        pitch, roll convention. Otherwise assume an nD rotation. Angles are
        assumed to be in degrees. They can be converted from radians with
        np.degrees if needed.
    shear : 1-D array or n-D array
        Either a vector of upper triangular values, or an nD shear matrix with
        ones along the main diagonal.
    affine : n-D array or napari.utils.transforms.Affine
        (N+1, N+1) affine transformation matrix in homogeneous coordinates.
        The first (N, N) entries correspond to a linear transform and
        the final column is a length N translation vector and a 1 or a napari
        `Affine` transform object. Applied as an extra transform on top of the
        provided scale, rotate, and shear values.
    opacity : float
        Opacity of the layer visual, between 0.0 and 1.0.
    blending : str
        One of a list of preset blending modes that determines how RGB and
        alpha values of the layer visual get mixed. Allowed values are
        {'opaque', 'translucent', 'translucent_no_depth', 'additive', and 'minimum'}.
    visible : bool
        Whether the layer visual is currently being displayed.
    multiscale : bool
        Whether the data is multiscale or not. Multiscale data is
        represented by a list of data objects and should go from largest to
        smallest.

    Attributes
    ----------
    name : str
        Unique name of the layer.
    opacity : float
        Opacity of the layer visual, between 0.0 and 1.0.
    visible : bool
        Whether the layer visual is currently being displayed.
    blending : Blending
        Determines how RGB and alpha values get mixed.

        * ``Blending.OPAQUE``
          Allows for only the top layer to be visible and corresponds to
          ``depth_test=True``, ``cull_face=False``, ``blend=False``.
        * ``Blending.TRANSLUCENT``
          Allows for multiple layers to be blended with different opacity and
          corresponds to ``depth_test=True``, ``cull_face=False``,
          ``blend=True``, ``blend_func=('src_alpha', 'one_minus_src_alpha')``,
          and ``blend_equation=('func_add')``.
        * ``Blending.TRANSLUCENT_NO_DEPTH``
          Allows for multiple layers to be blended with different opacity, but
          no depth testing is performed. Corresponds to ``depth_test=False``,
          ``cull_face=False``, ``blend=True``,
          ``blend_func=('src_alpha', 'one_minus_src_alpha')``, and
          ``blend_equation=('func_add')``.
        * ``Blending.ADDITIVE``
          Allows for multiple layers to be blended together with different
          colors and opacity. Useful for creating overlays. It corresponds to
          ``depth_test=False``, ``cull_face=False``, ``blend=True``,
          ``blend_func=('src_alpha', 'one')``, and ``blend_equation=('func_add')``.
        * ``Blending.MINIMUM``
            Allows for multiple layers to be blended together such that
            the minimum of each RGB component and alpha are selected.
            Useful for creating overlays with inverted colormaps. It
            corresponds to ``depth_test=False``, ``cull_face=False``, ``blend=True``,
            ``blend_equation=('min')``.
    scale : tuple of float
        Scale factors for the layer.
    translate : tuple of float
        Translation values for the layer.
    rotate : float, 3-tuple of float, or n-D array.
        If a float convert into a 2D rotation matrix using that value as an
        angle. If 3-tuple convert into a 3D rotation matrix, using a yaw,
        pitch, roll convention. Otherwise assume an nD rotation. Angles are
        assumed to be in degrees. They can be converted from radians with
        np.degrees if needed.
    shear : 1-D array or n-D array
        Either a vector of upper triangular values, or an nD shear matrix with
        ones along the main diagonal.
    affine : n-D array or napari.utils.transforms.Affine
        (N+1, N+1) affine transformation matrix in homogeneous coordinates.
        The first (N, N) entries correspond to a linear transform and
        the final column is a length N translation vector and a 1 or a napari
        `Affine` transform object. Applied as an extra transform on top of the
        provided scale, rotate, and shear values.
    multiscale : bool
        Whether the data is multiscale or not. Multiscale data is
        represented by a list of data objects and should go from largest to
        smallest.
    cache : bool
        Whether slices of out-of-core datasets should be cached upon retrieval.
        Currently, this only applies to dask arrays.
    z_index : int
        Depth of the layer visual relative to other visuals in the scenecanvas.
    corner_pixels : array
        Coordinates of the top-left and bottom-right canvas pixels in the data
        coordinates of each layer. For multiscale data the coordinates are in
        the space of the currently viewed data level, not the highest resolution
        level.
    ndim : int
        Dimensionality of the layer.
    thumbnail : (N, M, 4) array
        Array of thumbnail data for the layer.
    status : str
        Displayed in status bar bottom left.
    help : str
        Displayed in status bar bottom right.
    interactive : bool
        Determine if canvas pan/zoom interactivity is enabled.
    cursor : str
        String identifying which cursor displayed over canvas.
    cursor_size : int | None
        Size of cursor if custom. None yields default size
    scale_factor : float
        Conversion factor from canvas coordinates to image coordinates, which
        depends on the current zoom level.
    source : Source
        source of the layer (such as a plugin or widget)

    Notes
    -----
    Must define the following:

    * `_extent_data`: property
    * `data` property (setter & getter)

    May define the following:

    * `_set_view_slice()`: called to set currently viewed slice
    * `_basename()`: base/default name of the layer
    """

    def __init__(
        self,
        data,
        ndim,
        *,
        name=None,
        metadata=None,
        scale=None,
        translate=None,
        rotate=None,
        shear=None,
        affine=None,
        opacity=1,
        blending='translucent',
        visible=True,
        multiscale=False,
        cache=True,  # this should move to future "data source" object.
        experimental_clipping_planes=None,
    ):
        super().__init__()

        if name is None and data is not None:
            name = magic_name(data)

        if scale is not None and not np.all(scale):
            raise ValueError(
                trans._(
                    "Layer {name} is invalid because it has scale values of 0. The layer's scale is currently {scale}",
                    deferred=True,
                    name=repr(name),
                    scale=repr(scale),
                )
            )

        # Needs to be imported here to avoid circular import in _source
        from .._source import current_source

        self._source = current_source()
        self.dask_optimized_slicing = configure_dask(data, cache)
        self._metadata = dict(metadata or {})
        self._opacity = opacity
        self._blending = Blending(blending)
        self._visible = visible
        self._freeze = False
        self._status = 'Ready'
        self._help = ''
        self._cursor = 'standard'
        self._cursor_size = 1
        self._interactive = True
        self._value = None
        self.scale_factor = 1
        self.multiscale = multiscale
        self._experimental_clipping_planes = ClippingPlaneList()

        self._ndim = ndim
        self._ndisplay = 2
        self._dims_order = list(range(ndim))

        # Create a transform chain consisting of four transforms:
        # 1. `tile2data`: An initial transform only needed to display tiles
        #   of an image. It maps pixels of the tile into the coordinate space
        #   of the full resolution data and can usually be represented by a
        #   scale factor and a translation. A common use case is viewing part
        #   of lower resolution level of a multiscale image, another is using a
        #   downsampled version of an image when the full image size is larger
        #   than the maximum allowed texture size of your graphics card.
        # 2. `data2physical`: The main transform mapping data to a world-like
        #   physical coordinate that may also encode acquisition parameters or
        #   sample spacing.
        # 3. `physical2world`: An extra transform applied in world-coordinates that
        #   typically aligns this layer with another.
        # 4. `world2grid`: An additional transform mapping world-coordinates
        #   into a grid for looking at layers side-by-side.
        if scale is None:
            scale = [1] * ndim
        if translate is None:
            translate = [0] * ndim
        self._transforms = TransformChain(
            [
                Affine(np.ones(ndim), np.zeros(ndim), name='tile2data'),
                CompositeAffine(
                    scale,
                    translate,
                    rotate=rotate,
                    shear=shear,
                    ndim=ndim,
                    name='data2physical',
                ),
                coerce_affine(affine, ndim=ndim, name='physical2world'),
                Affine(np.ones(ndim), np.zeros(ndim), name='world2grid'),
            ]
        )

        self._dims_point = [0] * ndim
        self.corner_pixels = np.zeros((2, ndim), dtype=int)
        self._editable = True
        self._array_like = False

        self._thumbnail_shape = (32, 32, 4)
        self._thumbnail = np.zeros(self._thumbnail_shape, dtype=np.uint8)
        self._update_properties = True
        self._name = ''
        self.experimental_clipping_planes = experimental_clipping_planes

        self.events = EmitterGroup(
            source=self,
            refresh=Event,
            set_data=Event,
            blending=Event,
            opacity=Event,
            visible=Event,
            scale=Event,
            translate=Event,
            rotate=Event,
            shear=Event,
            affine=Event,
            data=Event,
            name=Event,
            thumbnail=Event,
            status=Event,
            help=Event,
            interactive=Event,
            cursor=Event,
            cursor_size=Event,
            editable=Event,
            loaded=Event,
            _ndisplay=Event,
            select=WarningEmitter(
                trans._(
                    "'layer.events.select' is deprecated and will be removed in napari v0.4.9, use 'viewer.layers.selection.events.changed' instead, and inspect the 'added' attribute on the event.",
                    deferred=True,
                ),
                type='select',
            ),
            deselect=WarningEmitter(
                trans._(
                    "'layer.events.deselect' is deprecated and will be removed in napari v0.4.9, use 'viewer.layers.selection.events.changed' instead, and inspect the 'removed' attribute on the event.",
                    deferred=True,
                ),
                type='deselect',
            ),
        )
        self.name = name

    def __str__(self):
        """Return self.name."""
        return self.name

    def __repr__(self):
        cls = type(self)
        return f"<{cls.__name__} layer {repr(self.name)} at {hex(id(self))}>"

    def _mode_setter_helper(self, mode, Modeclass):
        """
        Helper to manage callbacks in multiple layers

        Parameters
        ----------
        mode : Modeclass | str
            New mode for the current layer.
        Modeclass : Enum
            Enum for the current class representing the modes it can takes,
            this is usually specific on each subclass.

        Returns
        -------
        tuple (new Mode, mode changed)

        """
        mode = Modeclass(mode)
        assert mode is not None
        if not self.editable:
            mode = Modeclass.PAN_ZOOM
        if mode == self._mode:
            return mode, False
        if mode.value not in Modeclass.keys():
            raise ValueError(
                trans._(
                    "Mode not recognized: {mode}", deferred=True, mode=mode
                )
            )
        old_mode = self._mode
        self._mode = mode

        for callback_list, mode_dict in [
            (self.mouse_drag_callbacks, self._drag_modes),
            (self.mouse_move_callbacks, self._move_modes),
            (
                self.mouse_double_click_callbacks,
                getattr(
                    self, '_double_click_modes', defaultdict(lambda: no_op)
                ),
            ),
        ]:
            if mode_dict[old_mode] in callback_list:
                callback_list.remove(mode_dict[old_mode])
            callback_list.append(mode_dict[mode])
        self.cursor = self._cursor_modes[mode]

        self.interactive = mode == Modeclass.PAN_ZOOM
        return mode, True

    @classmethod
    def _basename(cls):
        return f'{cls.__name__}'

    @property
    def name(self):
        """str: Unique name of the layer."""
        return self._name

    @name.setter
    def name(self, name):
        if name == self.name:
            return
        if not name:
            name = self._basename()
        self._name = str(name)
        self.events.name()

    @property
    def metadata(self) -> dict:
        """Key/value map for user-stored data."""
        return self._metadata

    @metadata.setter
    def metadata(self, value: dict) -> None:
        self._metadata.clear()
        self._metadata.update(value)

    @property
    def source(self):
        return self._source

    @property
    def loaded(self) -> bool:
        """Return True if this layer is fully loaded in memory.

        This base class says that layers are permanently in the loaded state.
        Derived classes that do asynchronous loading can override this.
        """
        return True

    @property
    def opacity(self):
        """float: Opacity value between 0.0 and 1.0."""
        return self._opacity

    @opacity.setter
    def opacity(self, opacity):
        if not 0.0 <= opacity <= 1.0:
            raise ValueError(
                trans._(
                    'opacity must be between 0.0 and 1.0; got {opacity}',
                    deferred=True,
                    opacity=opacity,
                )
            )

        self._opacity = opacity
        self._update_thumbnail()
        self.events.opacity()

    @property
    def blending(self):
        """Blending mode: Determines how RGB and alpha values get mixed.

        Blending.OPAQUE
            Allows for only the top layer to be visible and corresponds to
            depth_test=True, cull_face=False, blend=False.
        Blending.TRANSLUCENT
            Allows for multiple layers to be blended with different opacity
            and corresponds to depth_test=True, cull_face=False,
            blend=True, blend_func=('src_alpha', 'one_minus_src_alpha'),
            and blend_equation=('func_add').
        Blending.TRANSLUCENT_NO_DEPTH
          Allows for multiple layers to be blended with different opacity, but
          no depth testing is performed. Corresponds to ``depth_test=False``,
          cull_face=False, blend=True, blend_func=('src_alpha', 'one_minus_src_alpha'),
          and blend_equation=('func_add').
        Blending.ADDITIVE
            Allows for multiple layers to be blended together with
            different colors and opacity. Useful for creating overlays. It
            corresponds to depth_test=False, cull_face=False, blend=True,
            blend_func=('src_alpha', 'one'), and blend_equation=('func_add').
        Blending.MINIMUM
            Allows for multiple layers to be blended together such that
            the minimum of each RGB component and alpha are selected.
            Useful for creating overlays with inverted colormaps. It
            corresponds to depth_test=False, cull_face=False, blend=True,
            blend_equation=('min').
        """
        return str(self._blending)

    @blending.setter
    def blending(self, blending):
        self._blending = Blending(blending)
        self.events.blending()

    @property
    def visible(self):
        """bool: Whether the visual is currently being displayed."""
        return self._visible

    @visible.setter
    def visible(self, visibility):
        self._visible = visibility
        self.refresh()
        self.events.visible()
        self.editable = self._set_editable() if self.visible else False

    @property
    def editable(self):
        """bool: Whether the current layer data is editable from the viewer."""
        return self._editable

    @editable.setter
    def editable(self, editable):
        if self._editable == editable:
            return
        self._editable = editable
        self._set_editable(editable=editable)
        self.events.editable()

    @property
    def scale(self):
        """list: Anisotropy factors to scale data into world coordinates."""
        return self._transforms['data2physical'].scale

    @scale.setter
    def scale(self, scale):
        if scale is None:
            scale = [1] * self.ndim
        self._transforms['data2physical'].scale = np.array(scale)
        self._update_dims()
        self.events.scale()

    @property
    def translate(self):
        """list: Factors to shift the layer by in units of world coordinates."""
        return self._transforms['data2physical'].translate

    @translate.setter
    def translate(self, translate):
        self._transforms['data2physical'].translate = np.array(translate)
        self._update_dims()
        self.events.translate()

    @property
    def rotate(self):
        """array: Rotation matrix in world coordinates."""
        return self._transforms['data2physical'].rotate

    @rotate.setter
    def rotate(self, rotate):
        self._transforms['data2physical'].rotate = rotate
        self._update_dims()
        self.events.rotate()

    @property
    def shear(self):
        """array: Shear matrix in world coordinates."""
        return self._transforms['data2physical'].shear

    @shear.setter
    def shear(self, shear):
        self._transforms['data2physical'].shear = shear
        self._update_dims()
        self.events.shear()

    @property
    def affine(self):
        """napari.utils.transforms.Affine: Extra affine transform to go from physical to world coordinates."""
        return self._transforms['physical2world']

    @affine.setter
    def affine(self, affine):
        # Assignment by transform name is not supported by TransformChain and
        # EventedList, so use the integer index instead. For more details, see:
        # https://github.com/napari/napari/issues/3058
        self._transforms[2] = coerce_affine(
            affine, ndim=self.ndim, name='physical2world'
        )
        self._update_dims()
        self.events.affine()

    @property
    def translate_grid(self):
        warnings.warn(
            trans._(
                "translate_grid will become private in v0.4.14. See Layer.translate or Layer.data_to_world() instead.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return self._translate_grid

    @translate_grid.setter
    def translate_grid(self, translate_grid):
        warnings.warn(
            trans._(
                "translate_grid will become private in v0.4.14. See Layer.translate or Layer.data_to_world() instead.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        self._translate_grid = translate_grid

    @property
    def _translate_grid(self):
        """list: Factors to shift the layer by."""
        return self._transforms['world2grid'].translate

    @_translate_grid.setter
    def _translate_grid(self, translate_grid):
        if np.all(self._translate_grid == translate_grid):
            return
        self._transforms['world2grid'].translate = np.array(translate_grid)
        self.events.translate()

    @property
    def _is_moving(self):
        return self._private_is_moving

    @_is_moving.setter
    def _is_moving(self, value):
        assert value in (True, False)
        if value:
            assert self._moving_coordinates is not None
        self._private_is_moving = value

    @property
    def _dims_displayed(self):
        """To be removed displayed dimensions."""
        # Ultimately we aim to remove all slicing information from the layer
        # itself so that layers can be sliced in different ways for multiple
        # canvas. See https://github.com/napari/napari/pull/1919#issuecomment-738585093
        # for additional discussion.
        return self._dims_order[-self._ndisplay :]

    @property
    def _dims_not_displayed(self):
        """To be removed not displayed dimensions."""
        # Ultimately we aim to remove all slicing information from the layer
        # itself so that layers can be sliced in different ways for multiple
        # canvas. See https://github.com/napari/napari/pull/1919#issuecomment-738585093
        # for additional discussion.
        return self._dims_order[: -self._ndisplay]

    @property
    def _dims_displayed_order(self):
        """To be removed order of displayed dimensions."""
        # Ultimately we aim to remove all slicing information from the layer
        # itself so that layers can be sliced in different ways for multiple
        # canvas. See https://github.com/napari/napari/pull/1919#issuecomment-738585093
        # for additional discussion.
        displayed = self._dims_displayed
        # equivalent to: order = np.argsort(displayed)
        order = sorted(range(len(displayed)), key=lambda x: displayed[x])
        return tuple(order)

    def _update_dims(self, event=None):
        """Update the dims model and clear the extent cache.

        This function needs to be called whenever data or transform information
        changes, and should be called before events get emitted.
        """
        from ...components.dims import reorder_after_dim_reduction

        ndim = self._get_ndim()

        old_ndim = self._ndim
        if old_ndim > ndim:
            keep_axes = range(old_ndim - ndim, old_ndim)
            self._transforms = self._transforms.set_slice(keep_axes)
            self._dims_point = self._dims_point[-ndim:]
            self._dims_order = list(
                reorder_after_dim_reduction(self._dims_order[-ndim:])
            )
        elif old_ndim < ndim:
            new_axes = range(ndim - old_ndim)
            self._transforms = self._transforms.expand_dims(new_axes)
            self._dims_point = [0] * (ndim - old_ndim) + self._dims_point
            self._dims_order = list(range(ndim - old_ndim)) + [
                o + ndim - old_ndim for o in self._dims_order
            ]

        self._ndim = ndim
        if 'extent' in self.__dict__:
            del self.extent

        self.refresh()  # This call is need for invalidate cache of extent in LayerList. If you remove it pleas ad another workaround.

    @property
    @abstractmethod
    def data(self):
        # user writes own docstring
        raise NotImplementedError()

    @data.setter
    @abstractmethod
    def data(self, data):
        raise NotImplementedError()

    @property
    @abstractmethod
    def _extent_data(self) -> np.ndarray:
        """Extent of layer in data coordinates.

        Returns
        -------
        extent_data : array, shape (2, D)
        """
        raise NotImplementedError()

    @property
    def _extent_world(self) -> np.ndarray:
        """Range of layer in world coordinates.

        Returns
        -------
        extent_world : array, shape (2, D)
        """
        # Get full nD bounding box
        return get_extent_world(
            self._extent_data, self._data_to_world, self._array_like
        )

    @cached_property
    def extent(self) -> Extent:
        """Extent of layer in data and world coordinates."""
        extent_data = self._extent_data
        data_to_world = self._data_to_world
        extent_world = get_extent_world(
            extent_data, data_to_world, self._array_like
        )
        return Extent(
            data=extent_data,
            world=extent_world,
            step=abs(data_to_world.scale),
        )

    @property
    def _slice_indices(self):
        """(D, ) array: Slice indices in data coordinates."""

        if len(self._dims_not_displayed) == 0:
            # all dims are displayed dimensions
            return (slice(None),) * self.ndim

        if self.ndim > self._ndisplay:
            inv_transform = self._data_to_world.inverse
            # Subspace spanned by non displayed dimensions
            non_displayed_subspace = np.zeros(self.ndim)
            for d in self._dims_not_displayed:
                non_displayed_subspace[d] = 1
            # Map subspace through inverse transform, ignoring translation
            _inv_transform = Affine(
                ndim=self.ndim,
                linear_matrix=inv_transform.linear_matrix,
                translate=None,
            )
            mapped_nd_subspace = _inv_transform(non_displayed_subspace)
            # Look at displayed subspace
            displayed_mapped_subspace = (
                mapped_nd_subspace[d] for d in self._dims_displayed
            )
            # Check that displayed subspace is null
            if any(abs(v) > 1e-8 for v in displayed_mapped_subspace):
                warnings.warn(
                    trans._(
                        'Non-orthogonal slicing is being requested, but is not fully supported. Data is displayed without applying an out-of-slice rotation or shear component.',
                        deferred=True,
                    ),
                    category=UserWarning,
                )

        slice_inv_transform = inv_transform.set_slice(self._dims_not_displayed)
        world_pts = [self._dims_point[ax] for ax in self._dims_not_displayed]
        data_pts = slice_inv_transform(world_pts)
        if getattr(self, "_round_index", True):
            # A round is taken to convert these values to slicing integers
            data_pts = np.round(data_pts).astype(int)

        indices = [slice(None)] * self.ndim
        for i, ax in enumerate(self._dims_not_displayed):
            indices[ax] = data_pts[i]

        return tuple(indices)

    @abstractmethod
    def _get_ndim(self):
        raise NotImplementedError()

    def _set_editable(self, editable=None):
        if editable is None:
            self.editable = True

    def _get_base_state(self):
        """Get dictionary of attributes on base layer.

        Returns
        -------
        state : dict
            Dictionary of attributes on base layer.
        """
        base_dict = {
            'name': self.name,
            'metadata': self.metadata,
            'scale': list(self.scale),
            'translate': list(self.translate),
            'rotate': [list(r) for r in self.rotate],
            'shear': list(self.shear),
            'affine': self.affine.affine_matrix,
            'opacity': self.opacity,
            'blending': self.blending,
            'visible': self.visible,
            'experimental_clipping_planes': [
                plane.dict() for plane in self.experimental_clipping_planes
            ],
        }
        return base_dict

    @abstractmethod
    def _get_state(self):
        raise NotImplementedError()

    @property
    def _type_string(self):
        return self.__class__.__name__.lower()

    def as_layer_data_tuple(self):
        state = self._get_state()
        state.pop('data', None)
        return self.data, state, self._type_string

    @property
    def thumbnail(self):
        """array: Integer array of thumbnail for the layer"""
        return self._thumbnail

    @thumbnail.setter
    def thumbnail(self, thumbnail):
        if 0 in thumbnail.shape:
            thumbnail = np.zeros(self._thumbnail_shape, dtype=np.uint8)
        if thumbnail.dtype != np.uint8:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                thumbnail = convert_to_uint8(thumbnail)

        padding_needed = np.subtract(self._thumbnail_shape, thumbnail.shape)
        pad_amounts = [(p // 2, (p + 1) // 2) for p in padding_needed]
        thumbnail = np.pad(thumbnail, pad_amounts, mode='constant')

        # blend thumbnail with opaque black background
        background = np.zeros(self._thumbnail_shape, dtype=np.uint8)
        background[..., 3] = 255

        f_dest = thumbnail[..., 3][..., None] / 255
        f_source = 1 - f_dest
        thumbnail = thumbnail * f_dest + background * f_source

        self._thumbnail = thumbnail.astype(np.uint8)
        self.events.thumbnail()

    @property
    def ndim(self):
        """int: Number of dimensions in the data."""
        return self._ndim

    @property
    def help(self):
        """str: displayed in status bar bottom right."""
        return self._help

    @help.setter
    def help(self, help):
        if help == self.help:
            return
        self._help = help
        self.events.help(help=help)

    @property
    def interactive(self):
        """bool: Determine if canvas pan/zoom interactivity is enabled."""
        return self._interactive

    @interactive.setter
    def interactive(self, interactive):
        if interactive == self._interactive:
            return
        self._interactive = interactive
        self.events.interactive(interactive=interactive)

    @property
    def cursor(self):
        """str: String identifying cursor displayed over canvas."""
        return self._cursor

    @cursor.setter
    def cursor(self, cursor):
        if cursor == self.cursor:
            return
        self._cursor = cursor
        self.events.cursor(cursor=cursor)

    @property
    def cursor_size(self):
        """int | None: Size of cursor if custom. None yields default size."""
        return self._cursor_size

    @cursor_size.setter
    def cursor_size(self, cursor_size):
        if cursor_size == self.cursor_size:
            return
        self._cursor_size = cursor_size
        self.events.cursor_size(cursor_size=cursor_size)

    @property
    def experimental_clipping_planes(self):
        return self._experimental_clipping_planes

    @experimental_clipping_planes.setter
    def experimental_clipping_planes(
        self,
        value: Union[
            dict,
            ClippingPlane,
            List[Union[ClippingPlane, dict]],
            ClippingPlaneList,
        ],
    ):
        self._experimental_clipping_planes.clear()
        if value is None:
            return

        if isinstance(value, (ClippingPlane, dict)):
            value = [value]
        for new_plane in value:
            plane = ClippingPlane()
            plane.update(new_plane)
            self._experimental_clipping_planes.append(plane)

    def set_view_slice(self):
        with self.dask_optimized_slicing():
            self._set_view_slice()

    @abstractmethod
    def _set_view_slice(self):
        raise NotImplementedError()

    def _slice_dims(self, point=None, ndisplay=2, order=None):
        """Slice data with values from a global dims model.

        Note this will likely be moved off the base layer soon.

        Parameters
        ----------
        point : list
            Values of data to slice at in world coordinates.
        ndisplay : int
            Number of dimensions to be displayed.
        order : list of int
            Order of dimensions, where last `ndisplay` will be
            rendered in canvas.
        """
        if point is None:
            ndim = self.ndim
        else:
            ndim = len(point)

        if order is None:
            order = list(range(ndim))

        # adjust the order of the global dims based on the number of
        # dimensions that a layer has - for example a global order of
        # [2, 1, 0, 3] -> [0, 1] for a layer that only has two dimensions
        # or -> [1, 0, 2] for a layer with three as that corresponds to
        # the relative order of the last two and three dimensions
        # respectively
        order = self._world_to_data_dims_displayed(order, ndim_world=ndim)

        if point is None:
            point = [0] * ndim
            nd = min(self.ndim, ndisplay)
            for i in order[-nd:]:
                point[i] = slice(None)
        else:
            point = list(point)

        # If no slide data has changed, then do nothing
        offset = ndim - self.ndim
        if (
            np.all(order == self._dims_order)
            and ndisplay == self._ndisplay
            and np.all(point[offset:] == self._dims_point)
        ):
            return

        self._dims_order = order
        if self._ndisplay != ndisplay:
            self._ndisplay = ndisplay
            self.events._ndisplay()

        # Update the point values
        self._dims_point = point[offset:]
        self._update_dims()
        self._set_editable()

    @abstractmethod
    def _update_thumbnail(self):
        raise NotImplementedError()

    @abstractmethod
    def _get_value(self, position):
        """Value of the data at a position in data coordinates.

        Parameters
        ----------
        position : tuple
            Position in data coordinates.

        Returns
        -------
        value : tuple
            Value of the data.
        """
        raise NotImplementedError()

    def get_value(
        self,
        position: Tuple[float],
        *,
        view_direction: Optional[np.ndarray] = None,
        dims_displayed: Optional[List[int]] = None,
        world=False,
    ):
        """Value of the data at a position.

        If the layer is not visible, return None.

        Parameters
        ----------
        position : tuple of float
            Position in either data or world coordinates.
        view_direction : Optional[np.ndarray]
            A unit vector giving the direction of the ray in nD world coordinates.
            The default value is None.
        dims_displayed : Optional[List[int]]
            A list of the dimensions currently being displayed in the viewer.
            The default value is None.
        world : bool
            If True the position is taken to be in world coordinates
            and converted into data coordinates. False by default.

        Returns
        -------
        value : tuple, None
            Value of the data. If the layer is not visible return None.
        """
        if self.visible:
            if world:
                ndim_world = len(position)

                if dims_displayed is not None:
                    # convert the dims_displayed to the layer dims.This accounts
                    # for differences in the number of dimensions in the world
                    # dims versus the layer and for transpose and rolls.
                    dims_displayed = dims_displayed_world_to_layer(
                        dims_displayed,
                        ndim_world=ndim_world,
                        ndim_layer=self.ndim,
                    )
                position = self.world_to_data(position)

            if (dims_displayed is not None) and (view_direction is not None):
                if len(dims_displayed) == 2 or self.ndim == 2:
                    value = self._get_value(position=tuple(position))

                elif len(dims_displayed) == 3:
                    view_direction = self._world_to_data_ray(
                        list(view_direction)
                    )
                    start_point, end_point = self.get_ray_intersections(
                        position=position,
                        view_direction=view_direction,
                        dims_displayed=dims_displayed,
                        world=False,
                    )
                    value = self._get_value_3d(
                        start_point=start_point,
                        end_point=end_point,
                        dims_displayed=dims_displayed,
                    )
            else:
                value = self._get_value(position)

        else:
            value = None
        # This should be removed as soon as possible, it is still
        # used in Points and Shapes.
        self._value = value
        return value

    def _get_value_3d(
        self,
        start_point: np.ndarray,
        end_point: np.ndarray,
        dims_displayed: List[int],
    ) -> Union[float, int]:
        """Get the layer data value along a ray

        Parameters
        ----------
        start_point : np.ndarray
            The start position of the ray used to interrogate the data.
        end_point : np.ndarray
            The end position of the ray used to interrogate the data.
        dims_displayed : List[int]
            The indices of the dimensions currently displayed in the Viewer.

        Returns
        -------
        value
            The data value along the supplied ray.

        """
        return None

    def projected_distance_from_mouse_drag(
        self,
        start_position: np.ndarray,
        end_position: np.ndarray,
        view_direction: np.ndarray,
        vector: np.ndarray,
        dims_displayed: Union[List, np.ndarray],
    ):
        """Calculate the length of the projection of a line between two mouse
        clicks onto a vector (or array of vectors) in data coordinates.

        Parameters
        ----------
        start_position : np.ndarray
            Starting point of the drag vector in data coordinates
        end_position : np.ndarray
            End point of the drag vector in data coordinates
        view_direction : np.ndarray
            Vector defining the plane normal of the plane onto which the drag
            vector is projected.
        vector : np.ndarray
            (3,) unit vector or (n, 3) array thereof on which to project the drag
            vector from start_event to end_event. This argument is defined in data
            coordinates.
        dims_displayed : Union[List, np.ndarray]
            (3,) list of currently displayed dimensions

        Returns
        -------
        projected_distance : (1, ) or (n, ) np.ndarray of float
        """
        start_position = self._world_to_displayed_data(
            start_position, dims_displayed
        )
        end_position = self._world_to_displayed_data(
            end_position, dims_displayed
        )
        view_direction = self._world_to_displayed_data_ray(
            view_direction, dims_displayed
        )
        return drag_data_to_projected_distance(
            start_position, end_position, view_direction, vector
        )

    @contextmanager
    def block_update_properties(self):
        previous = self._update_properties
        self._update_properties = False
        try:
            yield
        finally:
            self._update_properties = previous

    def _set_highlight(self, force=False):
        """Render layer highlights when appropriate.

        Parameters
        ----------
        force : bool
            Bool that forces a redraw to occur when `True`.
        """
        pass

    def refresh(self, event=None):
        """Refresh all layer data based on current view slice."""
        if self.visible:
            self.set_view_slice()
            self.events.set_data()  # refresh is called in _update_dims which means that extent cache is invalidated. Then, base on this event extent cache in layerlist is invalidated.
            self._update_thumbnail()
            self._set_highlight(force=True)

    def world_to_data(self, position):
        """Convert from world coordinates to data coordinates.

        Parameters
        ----------
        position : tuple, list, 1D array
            Position in world coordinates. If longer then the
            number of dimensions of the layer, the later
            dimensions will be used.

        Returns
        -------
        tuple
            Position in data coordinates.
        """
        if len(position) >= self.ndim:
            coords = list(position[-self.ndim :])
        else:
            coords = [0] * (self.ndim - len(position)) + list(position)

        return tuple(self._transforms[1:].simplified.inverse(coords))

    def data_to_world(self, position):
        """Convert from data coordinates to world coordinates.

        Parameters
        ----------
        position : tuple, list, 1D array
            Position in data coordinates. If longer then the
            number of dimensions of the layer, the later
            dimensions will be used.

        Returns
        -------
        tuple
            Position in world coordinates.
        """
        if len(position) >= self.ndim:
            coords = list(position[-self.ndim :])
        else:
            coords = [0] * (self.ndim - len(position)) + list(position)

        return tuple(self._transforms[1:].simplified(coords))

    def _world_to_displayed_data(
        self, position: np.ndarray, dims_displayed: np.ndarray
    ) -> tuple:
        """Convert world to data coordinates for displayed dimensions only.

        Parameters
        ----------
        position : tuple, list, 1D array
            Position in world coordinates. If longer then the
            number of dimensions of the layer, the later
            dimensions will be used.
        dims_displayed : list, 1D array
            Indices of displayed dimensions of the data.

        Returns
        -------
        tuple
            Position in data coordinates for the displayed dimensions only
        """
        position_nd = self.world_to_data(position)
        position_ndisplay = np.asarray(position_nd)[dims_displayed]
        return tuple(position_ndisplay)

    @property
    def _data_to_world(self) -> Affine:
        """The transform from data to world coordinates.

        This affine transform is composed from the affine property and the
        other transform properties in the following order:

        affine * (rotate * shear * scale + translate)
        """
        return self._transforms[1:3].simplified

    def _world_to_data_ray(self, vector) -> tuple:
        """Convert a vector defining an orientation from world coordinates to data coordinates.
        For example, this would be used to convert the view ray.

        Parameters
        ----------
        vector : tuple, list, 1D array
            A vector in world coordinates.

        Returns
        -------
        tuple
            Vector in data coordinates.
        """
        p1 = np.asarray(self.world_to_data(vector))
        p0 = np.asarray(self.world_to_data(np.zeros_like(vector)))
        normalized_vector = (p1 - p0) / np.linalg.norm(p1 - p0)

        return tuple(normalized_vector)

    def _world_to_displayed_data_ray(
        self, vector_world, dims_displayed
    ) -> np.ndarray:
        """Convert an orientation from world to displayed data coordinates.

        For example, this would be used to convert the view ray.

        Parameters
        ----------
        vector_world : tuple, list, 1D array
            A vector in world coordinates.

        Returns
        -------
        tuple
            Vector in data coordinates.
        """
        vector_data_nd = np.asarray(self._world_to_data_ray(vector_world))
        vector_data_ndisplay = vector_data_nd[dims_displayed]
        vector_data_ndisplay /= np.linalg.norm(vector_data_ndisplay)
        return vector_data_ndisplay

    def _world_to_data_dims_displayed(
        self, dims_displayed: List[int], ndim_world: int
    ) -> List[int]:
        """Convert indices of displayed dims from world to data coordinates.

        This accounts for differences in dimensionality between the world
        and the data coordinates. For example a world dims order of
        [2, 1, 0, 3] would be [0, 1] for a layer that only has two dimensions
        or [1, 0, 2] for a layer with three as that corresponds to the
        relative order of the last two and three dimensions respectively

        Parameters
        ----------
        dims_displayed : List[int]
            The world displayed dimensions.
        ndim_world : int
            The number of dimensions in the world coordinate system.

        Returns
        -------
        dims_displayed_data : List[int]
            The displayed dimensions in data coordinates.
        """
        offset = ndim_world - self.ndim
        order = np.array(dims_displayed)
        if offset <= 0:
            return list(range(-offset)) + list(order - offset)
        else:
            return list(order[order >= offset] - offset)

    def _display_bounding_box(self, dims_displayed: np.ndarray):
        """An axis aligned (self._ndisplay, 2) bounding box around the data"""
        return self._extent_data[:, dims_displayed].T

    def click_plane_from_click_data(
        self,
        click_position: np.ndarray,
        view_direction: np.ndarray,
        dims_displayed: List,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Calculate a (point, normal) plane parallel to the canvas in data
        coordinates, centered on the centre of rotation of the camera.

        Parameters
        ----------
        click_position : np.ndarray
            click position in world coordinates from mouse event.
        view_direction : np.ndarray
            view direction in world coordinates from mouse event.
        dims_displayed : List
            dimensions of the data array currently in view.

        Returns
        -------
        click_plane : Tuple[np.ndarray, np.ndarray]
            tuple of (plane_position, plane_normal) in data coordinates.
        """
        click_position = np.asarray(click_position)
        view_direction = np.asarray(view_direction)
        plane_position = self.world_to_data(click_position)[dims_displayed]
        plane_normal = self._world_to_data_ray(view_direction)[dims_displayed]
        return plane_position, plane_normal

    def get_ray_intersections(
        self,
        position: List[float],
        view_direction: np.ndarray,
        dims_displayed: List[int],
        world: bool = True,
    ) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[None, None]]:
        """Get the start and end point for the ray extending
        from a point through the data bounding box.

        Parameters
        ----------
        position
            the position of the point in nD coordinates. World vs. data
            is set by the world keyword argument.
        view_direction : np.ndarray
            a unit vector giving the direction of the ray in nD coordinates.
            World vs. data is set by the world keyword argument.
        dims_displayed
            a list of the dimensions currently being displayed in the viewer.
        world : bool
            True if the provided coordinates are in world coordinates.
            Default value is True.

        Returns
        -------
        start_point : np.ndarray
            The point on the axis-aligned data bounding box that the cursor click
            intersects with. This is the point closest to the camera.
            The point is the full nD coordinates of the layer data.
            If the click does not intersect the axis-aligned data bounding box,
            None is returned.
        end_point : np.ndarray
            The point on the axis-aligned data bounding box that the cursor click
            intersects with. This is the point farthest from the camera.
            The point is the full nD coordinates of the layer data.
            If the click does not intersect the axis-aligned data bounding box,
            None is returned.
        """
        if len(dims_displayed) != 3:
            return None, None

        # create the bounding box in data coordinates
        bounding_box = self._display_bounding_box(dims_displayed)

        start_point, end_point = self._get_ray_intersections(
            position=position,
            view_direction=view_direction,
            dims_displayed=dims_displayed,
            world=world,
            bounding_box=bounding_box,
        )
        return start_point, end_point

    def _get_offset_data_position(self, position: List[float]) -> List[float]:
        """Adjust position for offset between viewer and data coordinates."""
        return position

    def _get_ray_intersections(
        self,
        position: List[float],
        view_direction: np.ndarray,
        dims_displayed: List[int],
        world: bool = True,
        bounding_box: Optional[np.ndarray] = None,
    ) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[None, None]]:
        """Get the start and end point for the ray extending
        from a point through the data bounding box.

        Parameters
        ----------
        position
            the position of the point in nD coordinates. World vs. data
            is set by the world keyword argument.
        view_direction : np.ndarray
            a unit vector giving the direction of the ray in nD coordinates.
            World vs. data is set by the world keyword argument.
        dims_displayed
            a list of the dimensions currently being displayed in the viewer.
        world : bool
            True if the provided coordinates are in world coordinates.
            Default value is True.
        bounding_box : np.ndarray
            A (2, 3) bounding box around the data currently in view

        Returns
        -------
        start_point : np.ndarray
            The point on the axis-aligned data bounding box that the cursor click
            intersects with. This is the point closest to the camera.
            The point is the full nD coordinates of the layer data.
            If the click does not intersect the axis-aligned data bounding box,
            None is returned.
        end_point : np.ndarray
            The point on the axis-aligned data bounding box that the cursor click
            intersects with. This is the point farthest from the camera.
            The point is the full nD coordinates of the layer data.
            If the click does not intersect the axis-aligned data bounding box,
            None is returned."""
        # get the view direction and click position in data coords
        # for the displayed dimensions only
        if world is True:
            view_dir = self._world_to_displayed_data_ray(
                view_direction, dims_displayed
            )
            click_pos_data = self._world_to_displayed_data(
                position, dims_displayed
            )
        else:

            # adjust for any offset between viewer and data coordinates
            position = self._get_offset_data_position(position)

            view_dir = np.asarray(view_direction)[dims_displayed]
            click_pos_data = np.asarray(position)[dims_displayed]

        # Determine the front and back faces
        front_face_normal, back_face_normal = find_front_back_face(
            click_pos_data, bounding_box, view_dir
        )
        if front_face_normal is None and back_face_normal is None:
            # click does not intersect the data bounding box
            return None, None

        # Calculate ray-bounding box face intersections
        start_point_displayed_dimensions = (
            intersect_line_with_axis_aligned_bounding_box_3d(
                click_pos_data, view_dir, bounding_box, front_face_normal
            )
        )
        end_point_displayed_dimensions = (
            intersect_line_with_axis_aligned_bounding_box_3d(
                click_pos_data, view_dir, bounding_box, back_face_normal
            )
        )

        # add the coordinates for the axes not displayed
        start_point = np.asarray(position)
        start_point[dims_displayed] = start_point_displayed_dimensions
        end_point = np.asarray(position)
        end_point[dims_displayed] = end_point_displayed_dimensions

        return start_point, end_point

    @property
    def _displayed_axes(self):
        # assignment upfront to avoid repeated computation of properties
        _dims_displayed = self._dims_displayed
        _dims_displayed_order = self._dims_displayed_order
        displayed_axes = [_dims_displayed[i] for i in _dims_displayed_order]
        return displayed_axes

    @property
    def _corner_pixels_displayed(self):
        displayed_axes = self._displayed_axes
        corner_pixels_displayed = self.corner_pixels[:, displayed_axes]
        return corner_pixels_displayed

    def _update_draw(
        self, scale_factor, corner_pixels_displayed, shape_threshold
    ):
        """Update canvas scale and corner values on draw.

        For layer multiscale determining if a new resolution level or tile is
        required.

        Parameters
        ----------
        scale_factor : float
            Scale factor going from canvas to world coordinates.
        corner_pixels_displayed : array, shape (2, 2)
            Coordinates of the top-left and bottom-right canvas pixels in
            world coordinates.
        shape_threshold : tuple
            Requested shape of field of view in data coordinates.
        """
        self.scale_factor = scale_factor

        displayed_axes = self._displayed_axes
        # must adjust displayed_axes according to _dims_order
        displayed_axes = np.asarray(
            [self._dims_order[d] for d in displayed_axes]
        )
        # we need to compute all four corners to compute a complete,
        # data-aligned bounding box, because top-left/bottom-right may not
        # remain top-left and bottom-right after transformations.
        all_corners = list(itertools.product(*corner_pixels_displayed.T))
        # Note that we ignore the first transform which is tile2data
        data_corners = (
            self._transforms[1:]
            .simplified.set_slice(displayed_axes)
            .inverse(all_corners)
        )

        # find the maximal data-axis-aligned bounding box containing all four
        # canvas corners and round them to ints
        data_bbox = np.stack(
            [np.min(data_corners, axis=0), np.max(data_corners, axis=0)]
        )
        data_bbox_int = np.stack(
            [np.floor(data_bbox[0]), np.ceil(data_bbox[1])]
        ).astype(int)

        if self._ndisplay == 2 and self.multiscale:
            level, scaled_corners = compute_multiscale_level_and_corners(
                data_bbox_int,
                shape_threshold,
                self.downsample_factors[:, displayed_axes],
            )
            corners = np.zeros((2, self.ndim), dtype=int)
            # The corner_pixels attribute stores corners in the data
            # space of the selected level. Using the level's data
            # shape only works for images, but that's the only case we
            # handle now and downsample_factors is also only on image layers.
            max_coords = np.take(self.data[level].shape, displayed_axes)
            corners[:, displayed_axes] = np.clip(scaled_corners, 0, max_coords)
            display_shape = tuple(
                corners[1, displayed_axes] - corners[0, displayed_axes]
            )
            if any(s == 0 for s in display_shape):
                return
            if self.data_level != level or not np.all(
                self.corner_pixels == corners
            ):
                self._data_level = level
                self.corner_pixels = corners
                self.refresh()

        else:
            # The stored corner_pixels attribute must contain valid indices.
            displayed_extent = self.extent.data[:, displayed_axes]
            data_bbox_clipped = np.clip(
                data_bbox_int, displayed_extent[0], displayed_extent[1]
            )
            corners = np.zeros((2, self.ndim), dtype=int)
            corners[:, displayed_axes] = data_bbox_clipped
            self.corner_pixels = corners

    def _get_source_info(self):
        components = {}
        if self.source.reader_plugin:
            components['layer_base'] = os.path.basename(self.source.path or '')
            components['source_type'] = 'plugin'
            try:
                components['plugin'] = pm.get_manifest(
                    self.source.reader_plugin
                ).display_name
            except KeyError:
                components['plugin'] = self.source.reader_plugin
            return components

        elif self.source.sample:
            components['layer_base'] = self.name
            components['source_type'] = 'sample'
            try:
                components['plugin'] = pm.get_manifest(
                    self.source.sample[0]
                ).display_name
            except KeyError:
                components['plugin'] = self.source.sample[0]
            return components

        elif self.source.widget:
            components['layer_base'] = self.name
            components['source_type'] = 'widget'
            components['plugin'] = self.source.widget._function.__name__
            return components

        else:
            components['layer_base'] = self.name
            components['source_type'] = ''
            components['plugin'] = ''
            return components

    def get_source_str(self):

        source_info = self._get_source_info()

        return (
            source_info['layer_base']
            + ', '
            + source_info['source_type']
            + ' : '
            + source_info['plugin']
        )

    def get_status(
        self,
        position: Optional[Tuple[float, ...]] = None,
        *,
        view_direction: Optional[np.ndarray] = None,
        dims_displayed: Optional[List[int]] = None,
        world=False,
    ):
        """
        Status message information of the data at a coordinate position.

        Parameters
        ----------
        position : tuple of float
            Position in either data or world coordinates.
        view_direction : Optional[np.ndarray]
            A unit vector giving the direction of the ray in nD world coordinates.
            The default value is None.
        dims_displayed : Optional[List[int]]
            A list of the dimensions currently being displayed in the viewer.
            The default value is None.
        world : bool
            If True the position is taken to be in world coordinates
            and converted into data coordinates. False by default.

        Returns
        -------
        source_info : dict
            Dictionary containing a information that can be used as a status update.
        """
        if position is not None:
            value = self.get_value(
                position,
                view_direction=view_direction,
                dims_displayed=dims_displayed,
                world=world,
            )
        else:
            value = None

        source_info = self._get_source_info()
        source_info['coordinates'] = generate_layer_coords_status(
            position, value
        )
        return source_info

    def _get_tooltip_text(
        self,
        position,
        *,
        view_direction: Optional[np.ndarray] = None,
        dims_displayed: Optional[List[int]] = None,
        world: bool = False,
    ):
        """
        tooltip message of the data at a coordinate position.

        Parameters
        ----------
        position : tuple
            Position in either data or world coordinates.
        view_direction : Optional[np.ndarray]
            A unit vector giving the direction of the ray in nD world coordinates.
            The default value is None.
        dims_displayed : Optional[List[int]]
            A list of the dimensions currently being displayed in the viewer.
            The default value is None.
        world : bool
            If True the position is taken to be in world coordinates
            and converted into data coordinates. False by default.

        Returns
        -------
        msg : string
            String containing a message that can be used as a tooltip.
        """
        return ""

    def save(self, path: str, plugin: Optional[str] = None) -> List[str]:
        """Save this layer to ``path`` with default (or specified) plugin.

        Parameters
        ----------
        path : str
            A filepath, directory, or URL to open.  Extensions may be used to
            specify output format (provided a plugin is available for the
            requested format).
        plugin : str, optional
            Name of the plugin to use for saving. If ``None`` then all plugins
            corresponding to appropriate hook specification will be looped
            through to find the first one that can save the data.

        Returns
        -------
        list of str
            File paths of any files that were written.
        """
        from ...plugins.io import save_layers

        return save_layers(path, [self], plugin=plugin)

    def _on_selection(self, selected: bool):
        # This method is a temporary workaround to the fact that the Points
        # layer needs to know when its selection state changes so that it can
        # update the highlight state.  This, along with the events.select and
        # events.deselect emitters, (and the LayerList._on_selection_event
        # method) can be removed once highlighting logic has been removed from
        # the layer model.
        if selected:
            self.events.select()
        else:
            self.events.deselect()

    @classmethod
    def create(
        cls, data, meta: dict = None, layer_type: Optional[str] = None
    ) -> Layer:
        """Create layer from `data` of type `layer_type`.

        Primarily intended for usage by reader plugin hooks and creating a
        layer from an unwrapped layer data tuple.

        Parameters
        ----------
        data : Any
            Data in a format that is valid for the corresponding `layer_type`.
        meta : dict, optional
            Dict of keyword arguments that will be passed to the corresponding
            layer constructor.  If any keys in `meta` are not valid for the
            corresponding layer type, an exception will be raised.
        layer_type : str
            Type of layer to add. Must be the (case insensitive) name of a
            Layer subclass.  If not provided, the layer is assumed to
            be "image", unless data.dtype is one of (np.int32, np.uint32,
            np.int64, np.uint64), in which case it is assumed to be "labels".

        Raises
        ------
        ValueError
            If ``layer_type`` is not one of the recognized layer types.
        TypeError
            If any keyword arguments in ``meta`` are unexpected for the
            corresponding `add_*` method for this layer_type.

        Examples
        --------
        A typical use case might be to upack a tuple of layer data with a
        specified layer_type.

        >>> data = (
        ...     np.random.random((10, 2)) * 20,
        ...     {'face_color': 'blue'},
        ...     'points',
        ... )
        >>> Layer.create(*data)

        """
        from ... import layers
        from ..image._image_utils import guess_labels

        layer_type = (layer_type or '').lower()

        # assumes that big integer type arrays are likely labels.
        if not layer_type:
            layer_type = guess_labels(data)

        if layer_type not in layers.NAMES:
            raise ValueError(
                trans._(
                    "Unrecognized layer_type: '{layer_type}'. Must be one of: {layer_names}.",
                    deferred=True,
                    layer_type=layer_type,
                    layer_names=layers.NAMES,
                )
            )

        Cls = getattr(layers, layer_type.title())

        try:
            return Cls(data, **(meta or {}))
        except Exception as exc:
            if 'unexpected keyword argument' not in str(exc):
                raise exc

            bad_key = str(exc).split('keyword argument ')[-1]
            raise TypeError(
                trans._(
                    "_add_layer_from_data received an unexpected keyword argument ({bad_key}) for layer type {layer_type}",
                    deferred=True,
                    bad_key=bad_key,
                    layer_type=layer_type,
                )
            ) from exc


mgui.register_type(type_=List[Layer], return_callback=add_layers_to_viewer)



import platform
from unittest.mock import patch

import pytest

from napari._qt.qt_main_window import Window, _QtMainWindow
from napari.utils.theme import (
    _themes,
    get_theme,
    register_theme,
    unregister_theme,
)


def test_current_viewer(make_napari_viewer):
    """Test that we can retrieve the "current" viewer window easily.

    ... where "current" means it was the last viewer the user interacted with.
    """
    assert _QtMainWindow.current() is None

    # when we create a new viewer it becomes accessible at Viewer.current()
    v1 = make_napari_viewer(title='v1')
    assert _QtMainWindow._instances == [v1.window._qt_window]
    assert _QtMainWindow.current() == v1.window._qt_window

    v2 = make_napari_viewer(title='v2')
    assert _QtMainWindow._instances == [
        v1.window._qt_window,
        v2.window._qt_window,
    ]
    assert _QtMainWindow.current() == v2.window._qt_window

    # Viewer.current() will always give the most recently activated viewer.
    v1.window.activate()
    assert _QtMainWindow.current() == v1.window._qt_window
    v2.window.activate()
    assert _QtMainWindow.current() == v2.window._qt_window

    # The list remembers the z-order of previous viewers ...
    v2.close()
    assert _QtMainWindow.current() == v1.window._qt_window
    assert _QtMainWindow._instances == [v1.window._qt_window]

    # and when none are left, Viewer.current() becomes None again
    v1.close()
    assert _QtMainWindow._instances == []
    assert _QtMainWindow.current() is None


def test_set_geometry(make_napari_viewer):
    viewer = make_napari_viewer()
    values = (70, 70, 1000, 700)
    viewer.window.set_geometry(*values)
    assert viewer.window.geometry() == values


@patch.object(Window, "_update_theme_no_event")
@patch.object(Window, "_remove_theme")
@patch.object(Window, "_add_theme")
def test_update_theme(
    mock_add_theme,
    mock_remove_theme,
    mock_update_theme_no_event,
    make_napari_viewer,
):
    viewer = make_napari_viewer()

    blue = get_theme("dark", False)
    blue.name = "blue"
    register_theme("blue", blue)

    # triggered when theme was added
    mock_add_theme.assert_called()
    mock_remove_theme.assert_not_called()

    unregister_theme("blue")
    # triggered when theme was removed
    mock_remove_theme.assert_called()

    mock_update_theme_no_event.assert_not_called()
    viewer.theme = "light"
    theme = _themes["light"]
    theme.icon = "#FF0000"
    mock_update_theme_no_event.assert_called()


def test_lazy_console(make_napari_viewer):
    v = make_napari_viewer()
    assert v.window._qt_viewer._console is None
    v.update_console({"test": "test"})
    assert v.window._qt_viewer._console is None


@pytest.mark.skipif(
    platform.system() == "Darwin", reason="Cannot control menu bar on MacOS"
)
def test_menubar_shortcut(make_napari_viewer):
    v = make_napari_viewer()
    v.show()
    assert v.window.main_menu.isVisible()
    assert not v.window._main_menu_shortcut.isEnabled()
    v.window._toggle_menubar_visible()
    assert not v.window.main_menu.isVisible()
    assert v.window._main_menu_shortcut.isEnabled()



"""Automatically generate names.
"""
import inspect
import re
from collections import ChainMap

from .misc import ROOT_DIR, formatdoc

sep = ' '
start = 1

# Match integer between square brackets at end of string if after space
# or at beginning of string or just match end of string
numbered_patt = re.compile(r'((?<=\A\[)|(?<=\s\[))(?:\d+|)(?=\]$)|$')


def _inc_name_count_sub(match):
    count = match.group(0)

    try:
        count = int(count)
    except ValueError:  # not an int
        count = f'{sep}[{start}]'
    else:
        count = f'{count + 1}'

    return count


@formatdoc
def inc_name_count(name):
    """Increase a name's count matching `{numbered_patt}` by ``1``.

    If the name is not already numbered, append '{sep}[{start}]'.

    Parameters
    ----------
    name : str
        Original name.

    Returns
    -------
    incremented_name : str
        Numbered name incremented by ``1``.
    """
    return numbered_patt.sub(_inc_name_count_sub, name, count=1)


class CallerFrame:
    """
    Context manager to access the namespace in one of the upper caller frames.

    It is a context manager in order to be able to properly cleanup references
    to some frame objects after it is gone.

    Constructor takes a predicate taking a index and frame and returning whether
    to skip this frame and keep walking up the stack. The index starts at 1
    (caller frame), and increases.

    For example the following gives you the caller:
        - at least 5 Frames up
        - at most 42 Frames up
        - first one outside of Napari

        def skip_napari_frames(index, frame):
            if index < 5:
                return True
            if index > 42:
                return False
            return frame.f_globals.get("__name__", '').startswith('napari')

        with CallerFrame(skip_napari_frames) as c:
            print(c.namespace)

    This will be used for two things:
        - find the name of a value in caller frame.
        - capture local namespace of `napari.run()` when starting the qt-console

    For more complex logic you could use a callable that keep track of
    previous/state/frames, though be careful, the predicate is not guarantied to
    be called on all subsequents frames.

    """

    def __init__(self, skip_predicate):
        self.predicate = skip_predicate
        self.namespace = {}
        self.names = ()

    def __enter__(self):

        frame = inspect.currentframe().f_back
        try:
            # See issue #1635 regarding potential AttributeError
            # since frame could be None.
            # https://github.com/napari/napari/pull/1635
            if inspect.isframe(frame):
                frame = frame.f_back

            # Iterate frames while filename starts with path_prefix (part of Napari)
            n = 1
            while (
                inspect.isframe(frame)
                and inspect.isframe(frame.f_back)
                and inspect.iscode(frame.f_code)
                and (self.predicate(n, frame))
            ):
                n += 1
                frame = frame.f_back
            self.frame = frame
            if inspect.isframe(frame) and inspect.iscode(frame.f_code):
                self.namespace = ChainMap(frame.f_locals, frame.f_globals)
                self.names = (
                    *frame.f_code.co_varnames,
                    *frame.f_code.co_names,
                )

        finally:
            # We need to delete the frame explicitly according to the inspect
            # documentation for deterministic removal of the frame.
            # Otherwise, proper deletion is dependent on a cycle detector and
            # automatic garbage collection.
            # See handle_stackframe_without_leak example at the following URLs:
            # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
            # https://bugs.python.org/issue543148
            del frame

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del self.namespace
        del self.names


def magic_name(value, *, path_prefix=ROOT_DIR):
    """Fetch the name of the variable with the given value passed to the calling function.

    Parameters
    ----------
    value : any
        The value of the desired variable.
    path_prefix : absolute path-like, kwonly
        The path prefixes to ignore.

    Returns
    -------
    name : str or None
        Name of the variable, if found.
    """
    # Iterate frames while filename starts with path_prefix (part of Napari)
    with CallerFrame(
        lambda n, frame: frame.f_code.co_filename.startswith(path_prefix)
    ) as w:
        varmap = w.namespace
        names = w.names
        for name in names:
            if (
                name.isidentifier()
                and name in varmap
                and varmap[name] is value
            ):
                return name
        return None



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Optional, Tuple

from app_model.expressions import ContextKey

from ...utils._dtype import normalize_dtype
from ...utils.translations import trans
from ._context_keys import ContextNamespace

if TYPE_CHECKING:
    from numpy.typing import DTypeLike

    from ...layers import Layer
    from ...utils.events import Selection

    LayerSel = Selection[Layer]


def _len(s: LayerSel) -> int:
    return len(s)


def _all_linked(s: LayerSel) -> bool:
    from ...layers.utils._link_layers import layer_is_linked

    return bool(s and all(layer_is_linked(x) for x in s))


def _n_unselected_links(s: LayerSel) -> int:
    from ...layers.utils._link_layers import get_linked_layers

    return len(get_linked_layers(*s) - s)


def _is_rgb(s: LayerSel) -> bool:
    return getattr(s.active, "rgb", False)


def _only_img(s: LayerSel) -> bool:
    return bool(s and all(x._type_string == "image" for x in s))


def _n_selected_imgs(s: LayerSel) -> int:
    return sum(x._type_string == "image" for x in s)


def _only_labels(s: LayerSel) -> bool:
    return bool(s and all(x._type_string == "labels" for x in s))


def _n_selected_labels(s: LayerSel) -> int:
    return sum(x._type_string == "labels" for x in s)


def _only_points(s: LayerSel) -> bool:
    return bool(s and all(x._type_string == "points" for x in s))


def _n_selected_points(s: LayerSel) -> int:
    return sum(x._type_string == "labels" for x in s)


def _only_shapes(s: LayerSel) -> bool:
    return bool(s and all(x._type_string == "shapes" for x in s))


def _n_selected_shapes(s: LayerSel) -> int:
    return sum(x._type_string == "shapes" for x in s)


def _only_surface(s: LayerSel) -> bool:
    return bool(s and all(x._type_string == "surface" for x in s))


def _n_selected_surfaces(s: LayerSel) -> int:
    return sum(x._type_string == "surface" for x in s)


def _only_vectors(s: LayerSel) -> bool:
    return bool(s and all(x._type_string == "vectors" for x in s))


def _n_selected_vectors(s: LayerSel) -> int:
    return sum(x._type_string == "vectors" for x in s)


def _only_tracks(s: LayerSel) -> bool:
    return bool(s and all(x._type_string == "tracks" for x in s))


def _n_selected_tracks(s: LayerSel) -> int:
    return sum(x._type_string == "tracks" for x in s)


def _active_type(s: LayerSel) -> Optional[str]:
    return s.active._type_string if s.active else None


def _active_ndim(s: LayerSel) -> Optional[int]:
    return getattr(s.active.data, "ndim", None) if s.active else None


def _active_shape(s: LayerSel) -> Optional[Tuple[int, ...]]:
    return getattr(s.active.data, "shape", None) if s.active else None


def _same_shape(s: LayerSel) -> bool:
    return len({getattr(x.data, "shape", ()) for x in s}) == 1


def _active_dtype(s: LayerSel) -> DTypeLike:
    dtype = None
    if s.active:
        with contextlib.suppress(AttributeError):
            dtype = normalize_dtype(s.active.data.dtype).__name__
    return dtype


def _same_type(s: LayerSel) -> bool:
    return len({x._type_string for x in s}) == 1


class LayerListContextKeys(ContextNamespace['LayerSel']):
    """These are the available context keys relating to a LayerList.

    along with default value, a description, and a function to retrieve the
    current value from layers.selection
    """

    num_selected_layers = ContextKey(
        0,
        trans._("Number of currently selected layers."),
        _len,
    )
    num_selected_layers_linked = ContextKey(
        False,
        trans._("True when all selected layers are linked."),
        _all_linked,
    )
    num_unselected_linked_layers = ContextKey(
        0,
        trans._("Number of unselected layers linked to selected layer(s)."),
        _n_unselected_links,
    )
    active_layer_is_rgb = ContextKey(
        False,
        trans._("True when the active layer is RGB."),
        _is_rgb,
    )
    active_layer_type = ContextKey['LayerSel', Optional[str]](
        None,
        trans._(
            "Lowercase name of active layer type, or None of none active."
        ),
        _active_type,
    )
    # TODO: try to reduce these `num_selected_x_layers` to a single set of strings
    # or something... however, this would require that our context expressions
    # support Sets, tuples, lists, etc...  which they currently do not.
    num_selected_image_layers = ContextKey(
        0,
        trans._("Number of selected image layers."),
        _n_selected_imgs,
    )
    num_selected_labels_layers = ContextKey(
        0,
        trans._("Number of selected labels layers."),
        _n_selected_labels,
    )
    num_selected_points_layers = ContextKey(
        0,
        trans._("Number of selected points layers."),
        _n_selected_points,
    )
    num_selected_shapes_layers = ContextKey(
        0,
        trans._("Number of selected shapes layers."),
        _n_selected_shapes,
    )
    num_selected_surface_layers = ContextKey(
        0,
        trans._("Number of selected surface layers."),
        _n_selected_surfaces,
    )
    num_selected_vectors_layers = ContextKey(
        0,
        trans._("Number of selected vectors layers."),
        _n_selected_vectors,
    )
    num_selected_tracks_layers = ContextKey(
        0,
        trans._("Number of selected tracks layers."),
        _n_selected_tracks,
    )
    active_layer_ndim = ContextKey['LayerSel', Optional[int]](
        None,
        trans._(
            "Number of dimensions in the active layer, or `None` if nothing is active."
        ),
        _active_ndim,
    )
    active_layer_shape = ContextKey['LayerSel', Optional[Tuple[int, ...]]](
        (),
        trans._("Shape of the active layer, or `None` if nothing is active."),
        _active_shape,
    )
    active_layer_dtype = ContextKey(
        None,
        trans._("Dtype of the active layer, or `None` if nothing is active."),
        _active_dtype,
    )
    all_selected_layers_same_shape = ContextKey(
        False,
        trans._("True when all selected layers have the same shape."),
        _same_shape,
    )
    all_selected_layers_same_type = ContextKey(
        False,
        trans._("True when all selected layers are of the same type."),
        _same_type,
    )



import pytest

from napari._app_model import constants, get_app
from napari._qt._qapp_model import build_qmodel_menu


@pytest.mark.parametrize('menu_id', list(constants.MenuId))
def test_build_qmodel_menu(qtbot, menu_id):
    """Test that we can build qmenus for all registered menu IDs"""
    app = get_app()
    menu = build_qmodel_menu(menu_id)
    qtbot.addWidget(menu)
    assert len(menu.actions()) >= len(app.menus.get_menu(menu_id))



from __future__ import annotations

import napari

from ...layers.utils.interactivity_utils import (
    orient_plane_normal_around_cursor,
)
from ...layers.utils.layer_utils import register_layer_action
from ...utils.translations import trans
from ._image_constants import Mode
from .image import Image


def register_image_action(description: str, repeatable: bool = False):
    return register_layer_action(Image, description, repeatable)


@Image.bind_key('z')
@register_image_action(trans._('Orient plane normal along z-axis'))
def orient_plane_normal_along_z(layer: Image):
    orient_plane_normal_around_cursor(layer, plane_normal=(1, 0, 0))


@Image.bind_key('y')
@register_image_action(trans._('orient plane normal along y-axis'))
def orient_plane_normal_along_y(layer: Image):
    orient_plane_normal_around_cursor(layer, plane_normal=(0, 1, 0))


@Image.bind_key('x')
@register_image_action(trans._('orient plane normal along x-axis'))
def orient_plane_normal_along_x(layer: Image):
    orient_plane_normal_around_cursor(layer, plane_normal=(0, 0, 1))


@register_image_action(trans._('orient plane normal along view direction'))
def orient_plane_normal_along_view_direction(layer: Image):
    viewer = napari.viewer.current_viewer()
    if viewer.dims.ndisplay != 3:
        return
    layer.plane.normal = layer._world_to_displayed_data_ray(
        viewer.camera.view_direction, dims_displayed=[-3, -2, -1]
    )


@Image.bind_key('o')
def synchronise_plane_normal_with_view_direction(layer: Image):
    viewer = napari.viewer.current_viewer()
    if viewer.dims.ndisplay != 3:
        return

    def sync_plane_normal_with_view_direction(event=None):
        """Plane normal syncronisation mouse callback."""
        layer.plane.normal = layer._world_to_displayed_data_ray(
            viewer.camera.view_direction, [-3, -2, -1]
        )

    # update plane normal and add callback to mouse drag
    sync_plane_normal_with_view_direction()
    viewer.camera.events.angles.connect(sync_plane_normal_with_view_direction)
    yield
    # remove callback on key release
    viewer.camera.events.angles.disconnect(
        sync_plane_normal_with_view_direction
    )


@Image.bind_key('Space')
def hold_to_pan_zoom(layer):
    """Hold to pan and zoom in the viewer."""
    if layer._mode != Mode.PAN_ZOOM:
        # on key press
        prev_mode = layer.mode
        layer.mode = Mode.PAN_ZOOM

        yield

        # on key release
        layer.mode = prev_mode


@register_image_action(trans._('Transform'))
def activate_image_select_mode(layer):
    layer.mode = Mode.TRANSFORM


@register_image_action(trans._('Pan/zoom'))
def activate_image_pan_zoom_mode(layer):
    layer.mode = Mode.PAN_ZOOM



from typing import Tuple

import numpy as np
from vispy.scene.widgets.viewbox import ViewBox

from ...layers import (
    Image,
    Labels,
    Layer,
    Points,
    Shapes,
    Surface,
    Tracks,
    Vectors,
)
from ...utils.config import async_octree
from ...utils.translations import trans
from ..layers.base import VispyBaseLayer
from ..layers.image import VispyImageLayer
from ..layers.labels import VispyLabelsLayer
from ..layers.points import VispyPointsLayer
from ..layers.shapes import VispyShapesLayer
from ..layers.surface import VispySurfaceLayer
from ..layers.tracks import VispyTracksLayer
from ..layers.vectors import VispyVectorsLayer

layer_to_visual = {
    Image: VispyImageLayer,
    Labels: VispyLabelsLayer,
    Points: VispyPointsLayer,
    Shapes: VispyShapesLayer,
    Surface: VispySurfaceLayer,
    Vectors: VispyVectorsLayer,
    Tracks: VispyTracksLayer,
}


if async_octree:
    from ...layers.image.experimental.octree_image import _OctreeImageBase
    from ..experimental.vispy_tiled_image_layer import VispyTiledImageLayer

    # Insert _OctreeImageBase in front so it gets picked over plain Image.
    new_mapping = {_OctreeImageBase: VispyTiledImageLayer}
    new_mapping.update(layer_to_visual)
    layer_to_visual = new_mapping


def create_vispy_visual(layer: Layer) -> VispyBaseLayer:
    """Create vispy visual for a layer based on its layer type.

    Parameters
    ----------
    layer : napari.layers._base_layer.Layer
        Layer that needs its property widget created.

    Returns
    -------
    visual : vispy.scene.visuals.VisualNode
        Vispy visual node
    """
    for layer_type, visual_class in layer_to_visual.items():
        if isinstance(layer, layer_type):
            return visual_class(layer)

    raise TypeError(
        trans._(
            'Could not find VispyLayer for layer of type {dtype}',
            deferred=True,
            dtype=type(layer),
        )
    )


def get_view_direction_in_scene_coordinates(
    view: ViewBox,
    ndim: int,
    dims_displayed: Tuple[int],
) -> np.ndarray:
    """Calculate the unit vector pointing in the direction of the view.

    This is only for 3D viewing, so it returns None when
    len(dims_displayed) == 2.
    Adapted From:
    https://stackoverflow.com/questions/37877592/
        get-view-direction-relative-to-scene-in-vispy/37882984

    Parameters
    ----------
    view : vispy.scene.widgets.viewbox.ViewBox
        The vispy view box object to get the view direction from.
    ndim : int
        The number of dimensions in the full nD dims model.
        This is typically from viewer.dims.ndim
    dims_displayed : Tuple[int]
        The indices of the dims displayed in the viewer.
        This is typically from viewer.dims.displayed.

    Returns
    -------
    view_vector : np.ndarray
        Unit vector in the direction of the view in scene coordinates.
        Axes are ordered zyx. If the viewer is in 2D
        (i.e., len(dims_displayed) == 2), view_vector is None.
    """
    # only return a vector when viewing in 3D
    if len(dims_displayed) == 2:
        return None

    tform = view.scene.transform
    w, h = view.canvas.size

    # get a point at the center of the canvas
    # (homogeneous screen coords)
    screen_center = np.array([w / 2, h / 2, 0, 1])

    # find a point just in front of the center point
    # transform both to world coords and find the vector
    d1 = np.array([0, 0, 1, 0])
    point_in_front_of_screen_center = screen_center + d1
    p1 = tform.imap(point_in_front_of_screen_center)
    p0 = tform.imap(screen_center)
    d2 = p1 - p0

    # in 3D world coordinates
    d3 = d2[0:3]
    d4 = d3 / np.linalg.norm(d3)

    # data are ordered xyz on vispy Volume
    d4 = d4[[2, 1, 0]]
    view_dir_world = np.zeros((ndim,))
    for i, d in enumerate(dims_displayed):
        view_dir_world[d] = d4[i]

    return view_dir_world



"""napari.components provides the public-facing models for widgets
and other utilities that the user will be able to programmatically interact
with.

Classes
-------
Dims
    Current indices along each data dimension, together with which dimensions
    are being displayed, projected, sliced...
LayerList
    List of layers currently present in the viewer.
ViewerModel
    Data viewer displaying the currently rendered scene and
    layer-related controls.
"""

from .camera import Camera
from .dims import Dims
from .layerlist import LayerList

# Note that importing _viewer_key_bindings is needed as the Viewer gets
# decorated with keybindings during that process, but it is not directly needed
# by our users and so is deleted below
from . import _viewer_key_bindings  # isort:skip
from .viewer_model import ViewerModel  # isort:skip

del _viewer_key_bindings

__all__ = ['Camera', 'Dims', 'LayerList', 'ViewerModel']



from __future__ import annotations

from qtpy.QtWidgets import QLabel, QToolTip


class QtToolTipLabel(QLabel):
    """A QLabel that provides instant tooltips on mouser hover."""

    def enterEvent(self, event):
        """Override to show tooltips instantly."""
        if self.toolTip():
            pos = self.mapToGlobal(self.contentsRect().center())
            QToolTip.showText(pos, self.toolTip(), self)

        super().enterEvent(event)



from __future__ import annotations

from functools import lru_cache
from typing import Dict

from app_model import Application

from ._submenus import SUBMENUS
from .actions._layer_actions import LAYER_ACTIONS
from .injection._processors import PROCESSORS
from .injection._providers import PROVIDERS

APP_NAME = 'napari'


class NapariApplication(Application):
    def __init__(self) -> None:
        # raise_synchronous_exceptions means that commands triggered via
        # ``execute_command`` will immediately raise exceptions. Normally,
        # `execute_command` returns a Future object (which by definition does not
        # raise exceptions until requested).  While we could use that future to raise
        # exceptions with `.result()`, for now, raising immediately should
        # prevent any unexpected silent errors.  We can turn it off later if we
        # adopt asynchronous command execution.
        super().__init__(APP_NAME, raise_synchronous_exceptions=True)

        self.injection_store.namespace = _napari_names  # type: ignore [assignment]
        self.injection_store.register(
            providers=PROVIDERS, processors=PROCESSORS
        )

        for action in LAYER_ACTIONS:
            self.register_action(action)

        self.menus.append_menu_items(SUBMENUS)

    @classmethod
    def get_app(cls) -> NapariApplication:
        return Application.get_app(APP_NAME) or cls()


@lru_cache(maxsize=1)
def _napari_names() -> Dict[str, object]:
    """Napari names to inject into local namespace when evaluating type hints."""
    import napari
    from napari import components, layers, viewer

    def _public_types(module):
        return {
            name: val
            for name, val in vars(module).items()
            if not name.startswith('_')
            and isinstance(val, type)
            and getattr(val, '__module__', '_').startswith('napari')
        }

    return {
        'napari': napari,
        **_public_types(components),
        **_public_types(layers),
        **_public_types(viewer),
    }


def get_app() -> NapariApplication:
    """Get the Napari Application singleton."""
    return NapariApplication.get_app()



import platform
from functools import partial
from typing import TYPE_CHECKING

from ...settings import get_settings
from ...utils import config as async_config
from ...utils.translations import trans
from ._util import NapariMenu, populate_menu

if TYPE_CHECKING:
    from ..qt_main_window import Window


class ViewMenu(NapariMenu):
    def __init__(self, window: 'Window'):
        self._win = window
        super().__init__(trans._('&View'), window._qt_window)

        def _toggle_dict(text, name, prop):
            # helper func to make a Action dict for togglers
            obj = getattr(window._qt_viewer.viewer, name)
            return {
                'text': text,
                'slot': partial(setattr, obj, prop),
                'checkable': True,
                'checked': getattr(obj, prop),
                'check_on': getattr(obj.events, prop),
            }

        settings = get_settings()

        ACTIONS = [
            {
                'text': trans._('Toggle Full Screen'),
                'slot': window._toggle_fullscreen,
                'shortcut': 'Ctrl+F',
            },
            {
                'when': platform.system() != "Darwin",
                'text': trans._('Toggle Menubar Visibility'),
                'slot': window._toggle_menubar_visible,
                'shortcut': 'Ctrl+M',
                'statusTip': trans._('Hide Menubar'),
            },
            {
                'text': trans._('Toggle Play'),
                'slot': window._toggle_play,
                'shortcut': 'Ctrl+Alt+P',
            },
            {},
            {
                'when': async_config.async_octree,
                'text': trans._('Toggle Chunk Outlines'),
                'slot': window._qt_viewer._toggle_chunk_outlines,
                'shortcut': 'Ctrl+Alt+O',
            },
            {
                'menu': 'Axes',
                'items': [
                    _toggle_dict(trans._('Visible'), 'axes', 'visible'),
                    _toggle_dict(trans._('Colored'), 'axes', 'colored'),
                    _toggle_dict(trans._('Labels'), 'axes', 'labels'),
                    _toggle_dict(trans._('Dashed'), 'axes', 'dashed'),
                    _toggle_dict(trans._('Arrows'), 'axes', 'arrows'),
                ],
            },
            {
                'menu': 'Scale Bar',
                'items': [
                    _toggle_dict(trans._('Visible'), 'scale_bar', 'visible'),
                    _toggle_dict(trans._('Colored'), 'scale_bar', 'colored'),
                    _toggle_dict(trans._('Ticks'), 'scale_bar', 'ticks'),
                ],
            },
            {
                'text': trans._('Layer Tooltip visibility'),
                'slot': self._tooltip_visibility_toggle,
                'checkable': True,
                'checked': settings.appearance.layer_tooltip_visibility,
                'check_on': settings.appearance.events.layer_tooltip_visibility,
            },
            {
                'text': trans._('Activity Dock'),
                'slot': window._status_bar._toggle_activity_dock,
                'checkable': True,
                'checked': window._qt_window._activity_dialog.isVisible(),
                'check_on': window._status_bar._activity_item._activityBtn.toggled,
            },
            {},
        ]

        populate_menu(self, ACTIONS)

    def _tooltip_visibility_toggle(self, value):
        get_settings().appearance.layer_tooltip_visibility = value



from qtpy.QtCore import Qt
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QSplashScreen

from ..qt_event_loop import NAPARI_ICON_PATH, get_app


class NapariSplashScreen(QSplashScreen):
    def __init__(self, width=360):
        get_app()
        pm = QPixmap(NAPARI_ICON_PATH).scaled(
            width,
            width,
            Qt.AspectRatioMode.KeepAspectRatio,
            Qt.TransformationMode.SmoothTransformation,
        )
        super().__init__(pm)
        self.show()



import json
from itertools import cycle, islice

import numpy as np
import pytest
from pydantic import ValidationError

from napari.layers.utils.color_manager import ColorManager, ColorProperties
from napari.utils.colormaps.categorical_colormap import CategoricalColormap
from napari.utils.colormaps.standardize_color import transform_color


def _make_cycled_properties(values, length):
    """Helper function to make property values
    Parameters
    ----------
    values
        The values to be cycled.
    length : int
        The length of the resulting property array
    Returns
    -------
    cycled_properties : np.ndarray
        The property array comprising the cycled values.
    """
    cycled_properties = np.array(list(islice(cycle(values), 0, length)))
    return cycled_properties


def test_color_manager_empty():
    cm = ColorManager()
    np.testing.assert_allclose(cm.colors, np.empty((0, 4)))
    assert cm.color_mode == 'direct'


color_mapping = {0: np.array([1, 1, 1, 1]), 1: np.array([1, 0, 0, 1])}
fallback_colors = np.array([[1, 0, 0, 1], [0, 1, 0, 1]])
default_fallback_color = np.array([[1, 1, 1, 1]])
categorical_map = CategoricalColormap(
    colormap=color_mapping, fallback_color=fallback_colors
)


@pytest.mark.parametrize(
    'cat_cmap,expected',
    [
        ({'colormap': color_mapping}, (color_mapping, default_fallback_color)),
        (
            {'colormap': color_mapping, 'fallback_color': fallback_colors},
            (color_mapping, fallback_colors),
        ),
        ({'fallback_color': fallback_colors}, ({}, fallback_colors)),
        (color_mapping, (color_mapping, default_fallback_color)),
        (fallback_colors, ({}, fallback_colors)),
        (categorical_map, (color_mapping, fallback_colors)),
    ],
)
def test_categorical_colormap_from_dict(cat_cmap, expected):
    colors = np.array([[1, 1, 1, 1], [1, 0, 0, 1], [0, 0, 0, 1]])
    cm = ColorManager(
        colors=colors, categorical_colormap=cat_cmap, color_mode='direct'
    )
    np.testing.assert_equal(cm.categorical_colormap.colormap, expected[0])
    np.testing.assert_almost_equal(
        cm.categorical_colormap.fallback_color.values, expected[1]
    )


def test_invalid_categorical_colormap():
    colors = np.array([[1, 1, 1, 1], [1, 0, 0, 1], [0, 0, 0, 1]])
    invalid_cmap = 42
    with pytest.raises(ValidationError):
        _ = ColorManager(
            colors=colors,
            categorical_colormap=invalid_cmap,
            color_mode='direct',
        )


c_prop_dict = {
    'name': 'point_type',
    'values': np.array(['A', 'B', 'C']),
    'current_value': np.array(['C']),
}
c_prop_obj = ColorProperties(**c_prop_dict)


@pytest.mark.parametrize(
    'c_props,expected',
    [
        (None, None),
        ({}, None),
        (c_prop_obj, c_prop_obj),
        (c_prop_dict, c_prop_obj),
    ],
)
def test_color_properties_coercion(c_props, expected):
    colors = np.array([[1, 1, 1, 1], [1, 0, 0, 1], [0, 0, 0, 1]])
    cm = ColorManager(
        colors=colors, color_properties=c_props, color_mode='direct'
    )
    assert cm.color_properties == expected


wrong_type = ('prop_1', np.array([1, 2, 3]))
invalid_keys = {'values': np.array(['A', 'B', 'C'])}


@pytest.mark.parametrize('c_props', [wrong_type, invalid_keys])
def test_invalid_color_properties(c_props):

    colors = np.array([[1, 1, 1, 1], [1, 0, 0, 1], [0, 0, 0, 1]])
    with pytest.raises(ValidationError):
        _ = ColorManager(
            colors=colors, color_properties=c_props, color_mode='direct'
        )


@pytest.mark.parametrize(
    'curr_color,expected',
    [
        (None, np.array([0, 0, 0, 1])),
        ([], np.array([0, 0, 0, 1])),
        ('red', np.array([1, 0, 0, 1])),
        ([1, 0, 0, 1], np.array([1, 0, 0, 1])),
    ],
)
def test_current_color_coercion(curr_color, expected):
    colors = np.array([[1, 1, 1, 1], [1, 0, 0, 1], [0, 0, 0, 1]])
    cm = ColorManager(
        colors=colors, current_color=curr_color, color_mode='direct'
    )
    np.testing.assert_allclose(cm.current_color, expected)


color_str = ['red', 'red', 'red']
color_list = [[1, 0, 0, 1], [1, 0, 0, 1], [1, 0, 0, 1]]
color_arr = np.asarray(color_list)


@pytest.mark.parametrize('color', [color_str, color_list, color_arr])
def test_color_manager_direct(color):
    cm = ColorManager(colors=color, color_mode='direct')
    color_mode = cm.color_mode
    assert color_mode == 'direct'
    expected_colors = np.array([[1, 0, 0, 1], [1, 0, 0, 1], [1, 0, 0, 1]])
    np.testing.assert_allclose(cm.colors, expected_colors)
    np.testing.assert_allclose(cm.current_color, expected_colors[-1])

    # test adding a color
    new_color = [1, 1, 1, 1]
    cm._add(new_color)
    np.testing.assert_allclose(cm.colors[-1], new_color)

    # test removing colors
    cm._remove([0, 3])
    np.testing.assert_allclose(cm.colors, expected_colors[1:3])

    # test pasting colors
    paste_colors = np.array([[0, 0, 0, 1], [0, 0, 0, 1]])
    cm._paste(colors=paste_colors, properties={})
    post_paste_colors = np.vstack((expected_colors[1:3], paste_colors))
    np.testing.assert_allclose(cm.colors, post_paste_colors)

    # refreshing the colors in direct mode should have no effect
    cm._refresh_colors(properties={})
    np.testing.assert_allclose(cm.colors, post_paste_colors)


@pytest.mark.parametrize('color', [color_str, color_list, color_arr])
def test_set_color_direct(color):
    """Test setting the colors via the set_color method in direct mode"""
    # create an empty color manager
    cm = ColorManager()
    np.testing.assert_allclose(cm.colors, np.empty((0, 4)))
    assert cm.color_mode == 'direct'

    # set colors
    expected_colors = np.array([[1, 0, 0, 1], [1, 0, 0, 1], [1, 0, 0, 1]])
    cm._set_color(
        color, n_colors=len(color), properties={}, current_properties={}
    )
    np.testing.assert_almost_equal(cm.colors, expected_colors)


def test_continuous_colormap():
    # create ColorManager with a continuous colormap
    n_colors = 10
    properties = {
        'name': 'point_type',
        'values': _make_cycled_properties([0, 1.5], n_colors),
    }
    cm = ColorManager(
        color_properties=properties,
        continuous_colormap='gray',
        color_mode='colormap',
    )
    color_mode = cm.color_mode
    assert color_mode == 'colormap'
    color_array = transform_color(['black', 'white'] * int(n_colors / 2))
    colors = cm.colors.copy()
    np.testing.assert_allclose(colors, color_array)
    np.testing.assert_allclose(cm.current_color, [1, 1, 1, 1])

    # Add 2 color elements and test their color
    cm._add(0, n_colors=2)
    cm_colors = cm.colors
    assert len(cm_colors) == n_colors + 2
    np.testing.assert_allclose(
        cm_colors,
        np.vstack(
            (color_array, transform_color('black'), transform_color('black'))
        ),
    )

    # Check removing data adjusts colors correctly
    cm._remove({0, 2, 11})
    cm_colors_2 = cm.colors
    assert len(cm_colors_2) == (n_colors - 1)
    np.testing.assert_allclose(
        cm_colors_2,
        np.vstack((color_array[1], color_array[3:], transform_color('black'))),
    )

    # adjust the clims
    cm.contrast_limits = (0, 3)
    updated_colors = cm.colors
    np.testing.assert_allclose(updated_colors[-2], [0.5, 0.5, 0.5, 1])

    # first verify that prop value 0 is colored black
    current_colors = cm.colors
    np.testing.assert_allclose(current_colors[-1], [0, 0, 0, 1])

    # change the colormap
    new_colormap = 'gray_r'
    cm.continuous_colormap = new_colormap
    assert cm.continuous_colormap.name == new_colormap

    # the props valued 0 should now be white
    updated_colors = cm.colors
    np.testing.assert_allclose(updated_colors[-1], [1, 1, 1, 1])

    # test pasting values
    paste_props = {'point_type': np.array([0, 0])}
    paste_colors = np.array([[1, 1, 1, 1], [1, 1, 1, 1]])
    cm._paste(colors=paste_colors, properties=paste_props)
    np.testing.assert_allclose(cm.colors[-2:], paste_colors)


def test_set_color_colormap():
    # make an empty colormanager
    init_color_properties = {
        'name': 'point_type',
        'values': np.empty(0),
        'current_value': np.array([1.5]),
    }
    cm = ColorManager(
        color_properties=init_color_properties,
        continuous_colormap='gray',
        color_mode='colormap',
    )

    # use the set_color method to update the colors
    n_colors = 10
    updated_properties = {
        'point_type': _make_cycled_properties([0, 1.5], n_colors)
    }
    current_properties = {'point_type': np.array([1.5])}
    cm._set_color(
        color='point_type',
        n_colors=n_colors,
        properties=updated_properties,
        current_properties=current_properties,
    )
    color_array = transform_color(['black', 'white'] * int(n_colors / 2))
    np.testing.assert_allclose(cm.colors, color_array)


color_cycle_str = ['red', 'blue']
color_cycle_rgb = [[1, 0, 0], [0, 0, 1]]
color_cycle_rgba = [[1, 0, 0, 1], [0, 0, 1, 1]]


@pytest.mark.parametrize(
    "color_cycle",
    [color_cycle_str, color_cycle_rgb, color_cycle_rgba],
)
def test_color_cycle(color_cycle):
    """Test setting color with a color cycle list"""
    # create Points using list color cycle
    n_colors = 10
    properties = {
        'name': 'point_type',
        'values': _make_cycled_properties(['A', 'B'], n_colors),
    }
    cm = ColorManager(
        color_mode='cycle',
        color_properties=properties,
        categorical_colormap=color_cycle,
    )
    color_mode = cm.color_mode
    assert color_mode == 'cycle'
    color_array = transform_color(
        list(islice(cycle(color_cycle), 0, n_colors))
    )
    np.testing.assert_allclose(cm.colors, color_array)

    # Add 2 color elements and test their color
    cm._add('A', n_colors=2)
    cm_colors = cm.colors
    assert len(cm_colors) == n_colors + 2
    np.testing.assert_allclose(
        cm_colors,
        np.vstack(
            (color_array, transform_color('red'), transform_color('red'))
        ),
    )

    # Check removing data adjusts colors correctly
    cm._remove({0, 2, 11})
    cm_colors_2 = cm.colors
    assert len(cm_colors_2) == (n_colors - 1)
    np.testing.assert_allclose(
        cm_colors_2,
        np.vstack((color_array[1], color_array[3:], transform_color('red'))),
    )

    # update the colormap
    cm.categorical_colormap = ['black', 'white']

    # the first color should now be black
    np.testing.assert_allclose(cm.colors[0], [0, 0, 0, 1])

    # test pasting values
    paste_props = {'point_type': np.array(['B', 'B'])}
    paste_colors = np.array([[0, 0, 0, 1], [0, 0, 0, 1]])
    cm._paste(colors=paste_colors, properties=paste_props)
    np.testing.assert_allclose(cm.colors[-2:], paste_colors)


def test_set_color_cycle():
    # make an empty colormanager
    init_color_properties = {
        'name': 'point_type',
        'values': np.empty(0),
        'current_value': np.array(['A']),
    }
    cm = ColorManager(
        color_properties=init_color_properties,
        categorical_colormap=['black', 'white'],
        mode='cycle',
    )

    # use the set_color method to update the colors
    n_colors = 10
    updated_properties = {
        'point_type': _make_cycled_properties(['A', 'B'], n_colors)
    }
    current_properties = {'point_type': np.array(['B'])}
    cm._set_color(
        color='point_type',
        n_colors=n_colors,
        properties=updated_properties,
        current_properties=current_properties,
    )
    color_array = transform_color(['black', 'white'] * int(n_colors / 2))
    np.testing.assert_allclose(cm.colors, color_array)


@pytest.mark.parametrize('n_colors', [0, 1, 5])
def test_init_color_manager_direct(n_colors):
    color_manager = ColorManager._from_layer_kwargs(
        colors='red',
        properties={},
        n_colors=n_colors,
        continuous_colormap='viridis',
        contrast_limits=None,
        categorical_colormap=[[0, 0, 0, 1], [1, 1, 1, 1]],
    )

    assert len(color_manager.colors) == n_colors
    assert color_manager.color_mode == 'direct'
    np.testing.assert_array_almost_equal(
        color_manager.current_color, [1, 0, 0, 1]
    )
    if n_colors > 0:
        expected_colors = np.tile([1, 0, 0, 1], (n_colors, 1))
        np.testing.assert_array_almost_equal(
            color_manager.colors, expected_colors
        )
    # test that colormanager state can be saved and loaded
    cm_dict = color_manager.dict()
    color_manager_2 = ColorManager._from_layer_kwargs(
        colors=cm_dict, properties={}, n_colors=n_colors
    )
    assert color_manager == color_manager_2

    # test json serialization
    json_str = color_manager.json()
    cm_json_dict = json.loads(json_str)
    color_manager_3 = ColorManager._from_layer_kwargs(
        colors=cm_json_dict, properties={}, n_colors=n_colors
    )
    assert color_manager == color_manager_3


def test_init_color_manager_cycle():
    n_colors = 10
    color_cycle = [[0, 0, 0, 1], [1, 1, 1, 1]]
    properties = {'point_type': _make_cycled_properties(['A', 'B'], n_colors)}
    color_manager = ColorManager._from_layer_kwargs(
        colors='point_type',
        properties=properties,
        n_colors=n_colors,
        continuous_colormap='viridis',
        contrast_limits=None,
        categorical_colormap=color_cycle,
    )

    assert len(color_manager.colors) == n_colors
    assert color_manager.color_mode == 'cycle'
    color_array = transform_color(
        list(islice(cycle(color_cycle), 0, n_colors))
    )
    np.testing.assert_allclose(color_manager.colors, color_array)
    assert color_manager.color_properties.current_value == 'B'

    # test that colormanager state can be saved and loaded
    cm_dict = color_manager.dict()
    color_manager_2 = ColorManager._from_layer_kwargs(
        colors=cm_dict, properties=properties
    )
    assert color_manager == color_manager_2

    # test json serialization
    json_str = color_manager.json()
    cm_json_dict = json.loads(json_str)
    color_manager_3 = ColorManager._from_layer_kwargs(
        colors=cm_json_dict, properties={}, n_colors=n_colors
    )
    assert color_manager == color_manager_3


def test_init_color_manager_cycle_with_colors_dict():
    """Test initializing color cycle ColorManager from layer kwargs
    where the colors are given as a dictionary of ColorManager
    fields/values
    """
    n_colors = 10
    color_cycle = [[0, 0, 0, 1], [1, 1, 1, 1]]
    properties = {'point_type': _make_cycled_properties(['A', 'B'], n_colors)}
    colors_dict = {
        'color_properties': 'point_type',
        'color_mode': 'cycle',
        'categorical_colormap': color_cycle,
    }
    color_manager = ColorManager._from_layer_kwargs(
        colors=colors_dict,
        properties=properties,
        n_colors=n_colors,
        continuous_colormap='viridis',
    )
    assert len(color_manager.colors) == n_colors
    assert color_manager.color_mode == 'cycle'
    color_array = transform_color(
        list(islice(cycle(color_cycle), 0, n_colors))
    )
    np.testing.assert_allclose(color_manager.colors, color_array)
    assert color_manager.color_properties.current_value == 'B'
    assert color_manager.continuous_colormap.name == 'viridis'


def test_init_empty_color_manager_cycle():
    n_colors = 0
    color_cycle = [[0, 0, 0, 1], [1, 1, 1, 1]]
    properties = {'point_type': ['A', 'B']}
    color_manager = ColorManager._from_layer_kwargs(
        colors='point_type',
        properties=properties,
        n_colors=n_colors,
        continuous_colormap='viridis',
        contrast_limits=None,
        categorical_colormap=color_cycle,
    )

    assert len(color_manager.colors) == n_colors
    assert color_manager.color_mode == 'cycle'

    np.testing.assert_allclose(color_manager.current_color, [0, 0, 0, 1])
    assert color_manager.color_properties.current_value == 'A'

    color_manager._add()
    np.testing.assert_allclose(color_manager.colors, [[0, 0, 0, 1]])

    color_manager.color_properties.current_value = 'B'
    color_manager._add()
    np.testing.assert_allclose(
        color_manager.colors, [[0, 0, 0, 1], [1, 1, 1, 1]]
    )

    # test that colormanager state can be saved and loaded
    cm_dict = color_manager.dict()
    color_manager_2 = ColorManager._from_layer_kwargs(
        colors=cm_dict, properties=properties
    )
    assert color_manager == color_manager_2


def test_init_color_manager_colormap():
    n_colors = 10
    color_cycle = [[0, 0, 0, 1], [1, 1, 1, 1]]
    properties = {'point_type': _make_cycled_properties([0, 1.5], n_colors)}
    color_manager = ColorManager._from_layer_kwargs(
        colors='point_type',
        properties=properties,
        n_colors=n_colors,
        continuous_colormap='gray',
        contrast_limits=None,
        categorical_colormap=color_cycle,
    )

    assert len(color_manager.colors) == n_colors
    assert color_manager.color_mode == 'colormap'
    color_array = transform_color(['black', 'white'] * int(n_colors / 2))
    colors = color_manager.colors.copy()
    np.testing.assert_allclose(colors, color_array)
    np.testing.assert_allclose(color_manager.current_color, [1, 1, 1, 1])
    assert color_manager.color_properties.current_value == 1.5

    # test that colormanager state can be saved and loaded
    cm_dict = color_manager.dict()
    color_manager_2 = ColorManager._from_layer_kwargs(
        colors=cm_dict, properties=properties
    )
    assert color_manager == color_manager_2

    # test json serialization
    json_str = color_manager.json()
    cm_json_dict = json.loads(json_str)
    color_manager_3 = ColorManager._from_layer_kwargs(
        colors=cm_json_dict, properties={}, n_colors=n_colors
    )
    assert color_manager == color_manager_3


def test_init_color_manager_colormap_with_colors_dict():
    """Test initializing colormap ColorManager from layer kwargs
    where the colors are given as a dictionary of ColorManager
    fields/values
    """
    n_colors = 10
    color_cycle = [[0, 0, 0, 1], [1, 1, 1, 1]]
    properties = {'point_type': _make_cycled_properties([0, 1.5], n_colors)}
    colors_dict = {
        'color_properties': 'point_type',
        'color_mode': 'colormap',
        'categorical_colormap': color_cycle,
        'continuous_colormap': 'gray',
    }
    color_manager = ColorManager._from_layer_kwargs(
        colors=colors_dict, properties=properties, n_colors=n_colors
    )
    assert len(color_manager.colors) == n_colors
    assert color_manager.color_mode == 'colormap'
    color_array = transform_color(['black', 'white'] * int(n_colors / 2))
    colors = color_manager.colors.copy()
    np.testing.assert_allclose(colors, color_array)
    np.testing.assert_allclose(color_manager.current_color, [1, 1, 1, 1])
    assert color_manager.color_properties.current_value == 1.5
    assert color_manager.continuous_colormap.name == 'gray'


def test_init_empty_color_manager_colormap():
    n_colors = 0
    color_cycle = [[0, 0, 0, 1], [1, 1, 1, 1]]
    properties = {'point_type': [0]}
    color_manager = ColorManager._from_layer_kwargs(
        colors='point_type',
        properties=properties,
        n_colors=n_colors,
        color_mode='colormap',
        continuous_colormap='gray',
        contrast_limits=None,
        categorical_colormap=color_cycle,
    )

    assert len(color_manager.colors) == n_colors
    assert color_manager.color_mode == 'colormap'

    np.testing.assert_allclose(color_manager.current_color, [0, 0, 0, 1])
    assert color_manager.color_properties.current_value == 0

    color_manager._add()
    np.testing.assert_allclose(color_manager.colors, [[1, 1, 1, 1]])

    color_manager.color_properties.current_value = 1.5
    color_manager._add(update_clims=True)
    np.testing.assert_allclose(
        color_manager.colors, [[0, 0, 0, 1], [1, 1, 1, 1]]
    )

    # test that colormanager state can be saved and loaded
    cm_dict = color_manager.dict()
    color_manager_2 = ColorManager._from_layer_kwargs(
        colors=cm_dict, properties=properties
    )
    assert color_manager == color_manager_2


def test_color_manager_invalid_color_properties():
    """Passing an invalid property name for color_properties
    should raise a KeyError
    """
    n_colors = 10
    color_cycle = [[0, 0, 0, 1], [1, 1, 1, 1]]
    properties = {'point_type': _make_cycled_properties([0, 1.5], n_colors)}
    colors_dict = {
        'color_properties': 'not_point_type',
        'color_mode': 'colormap',
        'categorical_colormap': color_cycle,
        'continuous_colormap': 'gray',
    }
    with pytest.raises(KeyError):
        _ = ColorManager._from_layer_kwargs(
            colors=colors_dict, properties=properties, n_colors=n_colors
        )


def test_refresh_colors():
    # create ColorManager with a continuous colormap
    n_colors = 4
    properties = {
        'name': 'point_type',
        'values': _make_cycled_properties([0, 1.5], n_colors),
    }
    cm = ColorManager(
        color_properties=properties,
        continuous_colormap='gray',
        color_mode='colormap',
    )
    color_mode = cm.color_mode
    assert color_mode == 'colormap'
    color_array = transform_color(['black', 'white'] * int(n_colors / 2))
    colors = cm.colors.copy()
    np.testing.assert_allclose(colors, color_array)
    np.testing.assert_allclose(cm.current_color, [1, 1, 1, 1])

    # after refresh, the color should now be white. since we didn't
    # update the color mapping, the other values should remain
    # unchanged even though we added a value that extends the range
    # of values
    new_properties = {'point_type': properties['values']}
    new_properties['point_type'][0] = 3
    cm._refresh_colors(new_properties, update_color_mapping=False)
    new_colors = color_array.copy()
    new_colors[0] = [1, 1, 1, 1]
    np.testing.assert_allclose(cm.colors, new_colors)

    # now, refresh the colors, but update the mapping
    cm._refresh_colors(new_properties, update_color_mapping=True)
    refreshed_colors = [
        [1, 1, 1, 1],
        [0.5, 0.5, 0.5, 1],
        [0, 0, 0, 1],
        [0.5, 0.5, 0.5, 1],
    ]
    np.testing.assert_allclose(cm.colors, refreshed_colors)



from typing import Generator, Optional, Tuple

import pytest
from npe2 import PackageMetadata

from napari._qt.dialogs import qt_plugin_dialog


def _iter_napari_hub_or_pypi_plugin_info(
    conda_forge: bool = True,
) -> Generator[Tuple[Optional[PackageMetadata], bool], None, None]:
    """Mock the hub and pypi methods to collect available plugins.

    This will mock `napari.plugins.hub.iter_hub_plugin_info` for napari-hub,
    and `napari.plugins.pypi.iter_napari_plugin_info` for pypi.

    It will return two fake plugins that will populate the available plugins
    list (the bottom one). The first plugin will not be available on
    conda-forge so will be greyed out ("test-name-0"). The second plugin will
    be available on conda-forge so will be enabled ("test-name-1").
    """
    # This mock `base_data`` will be the same for both fake plugins.
    base_data = {
        "metadata_version": "1.0",
        "version": "0.1.0",
        "summary": "some test package",
        "home_page": "http://napari.org",
        "author": "test author",
        "license": "UNKNOWN",
    }
    for i in range(2):
        yield PackageMetadata(name=f"test-name-{i}", **base_data), bool(i)


@pytest.fixture
def plugin_dialog(qtbot, monkeypatch):
    """Fixture that provides a plugin dialog for a normal napari install."""
    for method_name in ["iter_hub_plugin_info", "iter_napari_plugin_info"]:
        monkeypatch.setattr(
            qt_plugin_dialog,
            method_name,
            _iter_napari_hub_or_pypi_plugin_info,
        )

    # This is patching `napari.utils.misc.running_as_constructor_app` function
    # to mock a normal napari install.
    monkeypatch.setattr(
        qt_plugin_dialog,
        "running_as_constructor_app",
        lambda: False,
    )

    widget = qt_plugin_dialog.QtPluginDialog()
    widget.show()
    qtbot.wait(300)
    qtbot.add_widget(widget)
    return widget


@pytest.fixture
def plugin_dialog_constructor(qtbot, monkeypatch):
    """
    Fixture that provides a plugin dialog for a constructor based install.
    """
    for method_name in ["iter_hub_plugin_info", "iter_napari_plugin_info"]:
        monkeypatch.setattr(
            qt_plugin_dialog,
            method_name,
            _iter_napari_hub_or_pypi_plugin_info,
        )

    # This is patching `napari.utils.misc.running_as_constructor_app` function
    # to mock a constructor based install.
    monkeypatch.setattr(
        qt_plugin_dialog,
        "running_as_constructor_app",
        lambda: True,
    )
    widget = qt_plugin_dialog.QtPluginDialog()
    widget.show()
    qtbot.wait(300)
    qtbot.add_widget(widget)
    return widget


def test_filter_not_available_plugins(plugin_dialog_constructor):
    """
    Check that the plugins listed under available plugins are
    enabled and disabled accordingly.

    The first plugin ("test-name-0") is not available on conda-forge and
    should be disabled, and show a tooltip warning.

    The second plugin ("test-name-1") is available on conda-forge and
    should be enabled without the tooltip warning.
    """
    item = plugin_dialog_constructor.available_list.item(0)
    widget = plugin_dialog_constructor.available_list.itemWidget(item)
    if widget:
        assert not widget.action_button.isEnabled()
        assert widget.warning_tooltip.isVisible()

    item = plugin_dialog_constructor.available_list.item(1)
    widget = plugin_dialog_constructor.available_list.itemWidget(item)
    assert widget.action_button.isEnabled()
    assert not widget.warning_tooltip.isVisible()


def test_filter_available_plugins(plugin_dialog):
    """
    Test the dialog is correctly filtering plugins in the available plugins
    list (the bottom one).
    """
    plugin_dialog.filter("")
    assert plugin_dialog.available_list.count() == 2
    assert plugin_dialog.available_list._count_visible() == 2

    plugin_dialog.filter("no-match@123")
    assert plugin_dialog.available_list._count_visible() == 0

    plugin_dialog.filter("")
    plugin_dialog.filter("test-name-0")
    assert plugin_dialog.available_list._count_visible() == 1


def test_filter_installed_plugins(plugin_dialog):
    """
    Test the dialog is correctly filtering plugins in the installed plugins
    list (the top one).
    """
    plugin_dialog.filter("")
    assert plugin_dialog.installed_list._count_visible() >= 0

    plugin_dialog.filter("no-match@123")
    assert plugin_dialog.installed_list._count_visible() == 0


def test_visible_widgets(plugin_dialog):
    """
    Test that the direct entry button and textbox are visible for
    normal napari installs.
    """

    assert plugin_dialog.direct_entry_edit.isVisible()
    assert plugin_dialog.direct_entry_btn.isVisible()


def test_constructor_visible_widgets(plugin_dialog_constructor):
    """
    Test that the direct entry button and textbox are hidden for
    constructor based napari installs.
    """
    assert not plugin_dialog_constructor.direct_entry_edit.isVisible()
    assert not plugin_dialog_constructor.direct_entry_btn.isVisible()



import warnings

from ...settings import *
from ..translations import trans

warnings.warn(
    trans._(
        "'napari.utils.settings' has moved to 'napari.settings' in 0.4.11. This will raise an ImportError in a future version",
        deferred=True,
    ),
    FutureWarning,
    stacklevel=2,
)
